"""Payloads for the Sky API requests.

All the payloads that will be used between the client and server communication
must be defined here to make sure it get covered by our API compatbility tests.

Compatibility note:
- Adding a new body for new API is compatible as long as the SDK method using
  the new API is properly decorated with `versions.minimal_api_version`.
- Adding a new field with default value to an existing body is compatible at
  API level, but the business logic must handle the case where the field is
  not proccessed by an old version of remote client/server. This can usually
  be done by checking `versions.get_remote_api_version()`.
- Other changes are not compatible at API level, so must be handled specially.
  A common pattern is to keep both the old and new version of the body and
  checking `versions.get_remote_api_version()` to decide which body to use. For
  example, say we refactor the `LaunchBody`, the original `LaunchBody` must be
  kept in the codebase and the new body should be added via `LaunchBodyV2`.
  Then if the remote runs in an old version, the local code should still send
  `LaunchBody` to keep the backward compatibility. `LaunchBody` can be removed
  later when constants.MIN_COMPATIBLE_API_VERSION is updated to a version that
  supports `LaunchBodyV2`

Also refer to sky.server.constants.MIN_COMPATIBLE_API_VERSION and the
sky.server.versions module for more details.
"""
import os
import typing
from typing import Any, Dict, List, Optional, Tuple, Union

from sky import admin_policy
from sky import serve
from sky import sky_logging
from sky import skypilot_config
from sky.adaptors import common as adaptors_common
from sky.server import common
from sky.skylet import autostop_lib
from sky.skylet import constants
from sky.usage import constants as usage_constants
from sky.usage import usage_lib
from sky.utils import annotations
from sky.utils import common as common_lib
from sky.utils import common_utils
from sky.utils import registry

if typing.TYPE_CHECKING:
    import pydantic
else:
    pydantic = adaptors_common.LazyImport('pydantic')

logger = sky_logging.init_logger(__name__)

# These non-skypilot environment variables will be updated from the local
# environment on each request when running a local API server.
# We should avoid adding variables here, but we should include credential-
# related variables.
EXTERNAL_LOCAL_ENV_VARS = [
    # Allow overriding the AWS authentication.
    'AWS_PROFILE',
    'AWS_DEFAULT_PROFILE',
    'AWS_ACCESS_KEY_ID',
    'AWS_SECRET_ACCESS_KEY',
    'AWS_SESSION_TOKEN',
    # Allow overriding the Azure authentication.
    'AZURE_CLIENT_ID',
    'AZURE_CLIENT_SECRET',
    'AZURE_TENANT_ID',
    'AZURE_SUBSCRIPTION_ID',
    # Allow overriding the GCP authentication.
    'GOOGLE_APPLICATION_CREDENTIALS',
    # Allow overriding the kubeconfig.
    'KUBECONFIG',
]


def request_body_env_vars() -> dict:
    env_vars = {}
    for env_var in os.environ:
        if (env_var.startswith(constants.SKYPILOT_ENV_VAR_PREFIX) and
                not env_var.startswith(
                    constants.SKYPILOT_SERVER_ENV_VAR_PREFIX)):
            env_vars[env_var] = os.environ[env_var]
        if common.is_api_server_local() and env_var in EXTERNAL_LOCAL_ENV_VARS:
            env_vars[env_var] = os.environ[env_var]
    env_vars[constants.USER_ID_ENV_VAR] = common_utils.get_user_hash()
    env_vars[constants.USER_ENV_VAR] = common_utils.get_local_user_name()
    env_vars[
        usage_constants.USAGE_RUN_ID_ENV_VAR] = usage_lib.messages.usage.run_id
    if not common.is_api_server_local():
        # Used in job controller, for local API server, keep the
        # SKYPILOT_CONFIG env var to use the config for the managed job.
        env_vars.pop(skypilot_config.ENV_VAR_SKYPILOT_CONFIG, None)
    # Remove the path to config file, as the config content is included in the
    # request body and will be merged with the config on the server side.
    env_vars.pop(skypilot_config.ENV_VAR_GLOBAL_CONFIG, None)
    env_vars.pop(skypilot_config.ENV_VAR_PROJECT_CONFIG, None)
    # Remove the config related env vars, as the client config override
    # should be passed in the request body.
    # Any new environment variables that are server-specific should
    # use SKYPILOT_SERVER_ENV_VAR_PREFIX.
    env_vars.pop(constants.ENV_VAR_DB_CONNECTION_URI, None)
    return env_vars


def get_override_skypilot_config_from_client() -> Dict[str, Any]:
    """Returns the override configs from the client."""
    if annotations.is_on_api_server:
        return {}
    config = skypilot_config.to_dict()
    # Remove the API server config, as we should not specify the SkyPilot
    # server endpoint on the server side. This avoids the warning at
    # server-side.
    config.pop_nested(('api_server',), default_value=None)
    # Remove the admin policy, as the policy has been applied on the client
    # side.
    config.pop_nested(('admin_policy',), default_value=None)
    return config


def get_override_skypilot_config_path_from_client() -> Optional[str]:
    """Returns the override config path from the client."""
    if annotations.is_on_api_server:
        return None
    # Currently, we don't need to check if the client-side config
    # has been overridden because we only deal with cases where
    # client has a project-level config/changed config and the
    # api server has a different config.
    return skypilot_config.loaded_config_path_serialized()


class BasePayload(pydantic.BaseModel):
    """The base payload for the SkyPilot API."""
    # Ignore extra fields in the request body, which is useful for backward
    # compatibility. The difference with `allow` is that `ignore` will not
    # include the unknown fields when dump the model, i.e., we can add new
    # fields to the request body without breaking the existing old API server
    # where the handler function does not accept the new field in function
    # signature.
    model_config = pydantic.ConfigDict(extra='ignore')


class RequestBody(BasePayload):
    """The request body for the SkyPilot API."""
    env_vars: Dict[str, str] = {}
    entrypoint: str = ''
    entrypoint_command: str = ''
    using_remote_api_server: bool = False
    override_skypilot_config: Optional[Dict[str, Any]] = {}
    override_skypilot_config_path: Optional[str] = None

    def __init__(self, **data):
        data['env_vars'] = data.get('env_vars', request_body_env_vars())
        usage_lib_entrypoint = usage_lib.messages.usage.entrypoint
        if usage_lib_entrypoint is None:
            usage_lib_entrypoint = ''
        data['entrypoint'] = data.get('entrypoint', usage_lib_entrypoint)
        data['entrypoint_command'] = data.get(
            'entrypoint_command', common_utils.get_pretty_entrypoint_cmd())
        data['using_remote_api_server'] = data.get(
            'using_remote_api_server', not common.is_api_server_local())
        data['override_skypilot_config'] = data.get(
            'override_skypilot_config',
            get_override_skypilot_config_from_client())
        data['override_skypilot_config_path'] = data.get(
            'override_skypilot_config_path',
            get_override_skypilot_config_path_from_client())
        super().__init__(**data)

    def to_kwargs(self) -> Dict[str, Any]:
        """Convert the request body to a kwargs dictionary on API server.

        This converts the request body into kwargs for the underlying SkyPilot
        backend's function.
        """
        kwargs = self.model_dump()
        kwargs.pop('env_vars')
        kwargs.pop('entrypoint')
        kwargs.pop('entrypoint_command')
        kwargs.pop('using_remote_api_server')
        kwargs.pop('override_skypilot_config')
        kwargs.pop('override_skypilot_config_path')
        return kwargs

    @property
    def user_hash(self) -> Optional[str]:
        return self.env_vars.get(constants.USER_ID_ENV_VAR)


class CheckBody(RequestBody):
    """The request body for the check endpoint."""
    clouds: Optional[Tuple[str, ...]] = None
    verbose: bool = False
    workspace: Optional[str] = None


class EnabledCloudsBody(RequestBody):
    """The request body for the enabled clouds endpoint."""
    workspace: Optional[str] = None
    expand: bool = False


class DagRequestBody(RequestBody):
    """Request body base class for endpoints with a dag."""
    dag: str

    def to_kwargs(self) -> Dict[str, Any]:
        # Import here to avoid requirement of the whole SkyPilot dependency on
        # local clients.
        # pylint: disable=import-outside-toplevel
        from sky.utils import dag_utils

        kwargs = super().to_kwargs()

        dag = dag_utils.load_chain_dag_from_yaml_str(self.dag)
        # We should not validate the dag here, as the file mounts are not
        # processed yet, but we need to validate the resources during the
        # optimization to make sure the resources are available.
        kwargs['dag'] = dag
        return kwargs


class DagRequestBodyWithRequestOptions(DagRequestBody):
    """Request body base class for endpoints with a dag and request options."""
    request_options: Optional[admin_policy.RequestOptions]

    def get_request_options(self) -> Optional[admin_policy.RequestOptions]:
        """Get the request options."""
        if self.request_options is None:
            return None
        if isinstance(self.request_options, dict):
            return admin_policy.RequestOptions(**self.request_options)
        return self.request_options

    def to_kwargs(self) -> Dict[str, Any]:
        kwargs = super().to_kwargs()
        kwargs['request_options'] = self.get_request_options()
        return kwargs


class ValidateBody(DagRequestBodyWithRequestOptions):
    """The request body for the validate endpoint."""
    dag: str


class OptimizeBody(DagRequestBodyWithRequestOptions):
    """The request body for the optimize endpoint."""
    dag: str
    minimize: common_lib.OptimizeTarget = common_lib.OptimizeTarget.COST


class LaunchBody(RequestBody):
    """The request body for the launch endpoint."""
    task: str
    cluster_name: str
    retry_until_up: bool = False
    # TODO(aylei): remove this field in v0.12.0
    idle_minutes_to_autostop: Optional[int] = None
    dryrun: bool = False
    # TODO(aylei): remove this field in v0.12.0
    down: bool = False
    backend: Optional[str] = None
    optimize_target: common_lib.OptimizeTarget = common_lib.OptimizeTarget.COST
    no_setup: bool = False
    clone_disk_from: Optional[str] = None
    fast: bool = False
    # Internal only:
    # pylint: disable=invalid-name
    quiet_optimizer: bool = False
    is_launched_by_jobs_controller: bool = False
    is_launched_by_sky_serve_controller: bool = False
    disable_controller_check: bool = False

    def to_kwargs(self) -> Dict[str, Any]:

        kwargs = super().to_kwargs()
        dag = common.process_mounts_in_task_on_api_server(self.task,
                                                          self.env_vars,
                                                          workdir_only=False)

        backend_cls = registry.BACKEND_REGISTRY.from_str(self.backend)
        backend = backend_cls() if backend_cls is not None else None
        kwargs['task'] = dag
        kwargs['backend'] = backend
        kwargs['_quiet_optimizer'] = kwargs.pop('quiet_optimizer')
        kwargs['_is_launched_by_jobs_controller'] = kwargs.pop(
            'is_launched_by_jobs_controller')
        kwargs['_is_launched_by_sky_serve_controller'] = kwargs.pop(
            'is_launched_by_sky_serve_controller')
        kwargs['_disable_controller_check'] = kwargs.pop(
            'disable_controller_check')
        return kwargs


class ExecBody(RequestBody):
    """The request body for the exec endpoint."""
    task: str
    cluster_name: str
    dryrun: bool = False
    down: bool = False
    backend: Optional[str] = None

    def to_kwargs(self) -> Dict[str, Any]:

        kwargs = super().to_kwargs()
        dag = common.process_mounts_in_task_on_api_server(self.task,
                                                          self.env_vars,
                                                          workdir_only=True)
        backend_cls = registry.BACKEND_REGISTRY.from_str(self.backend)
        backend = backend_cls() if backend_cls is not None else None
        kwargs['task'] = dag
        kwargs['backend'] = backend
        return kwargs


class StopOrDownBody(RequestBody):
    cluster_name: str
    purge: bool = False


class StatusBody(RequestBody):
    """The request body for the status endpoint."""
    cluster_names: Optional[List[str]] = None
    refresh: common_lib.StatusRefreshMode = common_lib.StatusRefreshMode.NONE
    all_users: bool = True
    # TODO (kyuds): default to False post 0.12.0
    include_credentials: bool = True
    # Only return fields that are needed for the
    # dashboard / CLI summary response
    summary_response: bool = False
    # Include the cluster handle in the response
    include_handle: bool = True


class StartBody(RequestBody):
    """The request body for the start endpoint."""
    cluster_name: str
    idle_minutes_to_autostop: Optional[int] = None
    wait_for: Optional[autostop_lib.AutostopWaitFor] = None
    retry_until_up: bool = False
    down: bool = False
    force: bool = False


class AutostopBody(RequestBody):
    """The request body for the autostop endpoint."""
    cluster_name: str
    idle_minutes: int
    wait_for: Optional[autostop_lib.AutostopWaitFor] = None
    down: bool = False


class QueueBody(RequestBody):
    """The request body for the queue endpoint."""
    cluster_name: str
    skip_finished: bool = False
    all_users: bool = False


class CancelBody(RequestBody):
    """The request body for the cancel endpoint."""
    cluster_name: str
    job_ids: Optional[List[int]]
    all: bool = False
    all_users: bool = False
    # Internal only. We cannot use prefix `_` because pydantic will not
    # include it in the request body.
    try_cancel_if_cluster_is_init: bool = False

    def to_kwargs(self) -> Dict[str, Any]:
        kwargs = super().to_kwargs()
        kwargs['_try_cancel_if_cluster_is_init'] = kwargs.pop(
            'try_cancel_if_cluster_is_init')
        return kwargs


class ProvisionLogsBody(RequestBody):
    """Cluster node."""
    cluster_name: str
    worker: Optional[int] = None


class ClusterJobBody(RequestBody):
    """The request body for the cluster job endpoint."""
    cluster_name: str
    job_id: Optional[int]
    follow: bool = True
    tail: int = 0


class ClusterJobsBody(RequestBody):
    """The request body for the cluster jobs endpoint."""
    cluster_name: str
    job_ids: Optional[List[str]]


class ClusterJobsDownloadLogsBody(RequestBody):
    """The request body for the cluster jobs download logs endpoint."""
    cluster_name: str
    job_ids: Optional[List[str]]
    local_dir: str = constants.SKY_LOGS_DIRECTORY


class UserCreateBody(RequestBody):
    """The request body for the user create endpoint."""
    username: str
    password: str
    role: Optional[str] = None


class UserDeleteBody(RequestBody):
    """The request body for the user delete endpoint."""
    user_id: str


class UserUpdateBody(RequestBody):
    """The request body for the user update endpoint."""
    user_id: str
    role: Optional[str] = None
    password: Optional[str] = None


class UserImportBody(RequestBody):
    """The request body for the user import endpoint."""
    csv_content: str


class ServiceAccountTokenCreateBody(RequestBody):
    """The request body for creating a service account token."""
    token_name: str
    expires_in_days: Optional[int] = None


class ServiceAccountTokenDeleteBody(RequestBody):
    """The request body for deleting a service account token."""
    token_id: str


class UpdateRoleBody(RequestBody):
    """The request body for updating a user role."""
    role: str


class ServiceAccountTokenRoleBody(RequestBody):
    """The request body for getting a service account token role."""
    token_id: str


class ServiceAccountTokenUpdateRoleBody(RequestBody):
    """The request body for updating a service account token role."""
    token_id: str
    role: str


class ServiceAccountTokenRotateBody(RequestBody):
    """The request body for rotating a service account token."""
    token_id: str
    expires_in_days: Optional[int] = None


class DownloadBody(RequestBody):
    """The request body for the download endpoint."""
    folder_paths: List[str]


class StorageBody(RequestBody):
    """The request body for the storage endpoint."""
    name: str


class VolumeApplyBody(RequestBody):
    """The request body for the volume apply endpoint."""
    name: str
    volume_type: str
    cloud: str
    region: Optional[str] = None
    zone: Optional[str] = None
    size: Optional[str] = None
    config: Optional[Dict[str, Any]] = None
    labels: Optional[Dict[str, str]] = None
    use_existing: Optional[bool] = None


class VolumeDeleteBody(RequestBody):
    """The request body for the volume delete endpoint."""
    names: List[str]


class VolumeListBody(RequestBody):
    """The request body for the volume list endpoint."""
    pass


class VolumeValidateBody(RequestBody):
    """The request body for the volume validate endpoint."""
    name: Optional[str] = None
    volume_type: Optional[str] = None
    infra: Optional[str] = None
    size: Optional[str] = None
    labels: Optional[Dict[str, str]] = None
    config: Optional[Dict[str, Any]] = None
    use_existing: Optional[bool] = None


class EndpointsBody(RequestBody):
    """The request body for the endpoint."""
    cluster: str
    port: Optional[Union[int, str]] = None


class ServeEndpointBody(RequestBody):
    """The request body for the serve controller endpoint."""
    port: Optional[Union[int, str]] = None


class JobStatusBody(RequestBody):
    """The request body for the job status endpoint."""
    cluster_name: str
    job_ids: Optional[List[int]]


class JobsLaunchBody(RequestBody):
    """The request body for the jobs launch endpoint."""
    task: str
    name: Optional[str]
    pool: Optional[str] = None
    num_jobs: Optional[int] = None

    def to_kwargs(self) -> Dict[str, Any]:
        kwargs = super().to_kwargs()
        kwargs['task'] = common.process_mounts_in_task_on_api_server(
            self.task, self.env_vars, workdir_only=False)
        return kwargs


class JobsQueueBody(RequestBody):
    """The request body for the jobs queue endpoint."""
    refresh: bool = False
    skip_finished: bool = False
    all_users: bool = False
    job_ids: Optional[List[int]] = None


class JobsQueueV2Body(RequestBody):
    """The request body for the jobs queue endpoint."""
    refresh: bool = False
    skip_finished: bool = False
    all_users: bool = False
    job_ids: Optional[List[int]] = None
    user_match: Optional[str] = None
    workspace_match: Optional[str] = None
    name_match: Optional[str] = None
    pool_match: Optional[str] = None
    page: Optional[int] = None
    limit: Optional[int] = None
    statuses: Optional[List[str]] = None
    # The fields to return in the response.
    # Refer to the fields in the `class ManagedJobRecord` in `response.py`
    fields: Optional[List[str]] = None


class JobsCancelBody(RequestBody):
    """The request body for the jobs cancel endpoint."""
    name: Optional[str] = None
    job_ids: Optional[List[int]] = None
    all: bool = False
    all_users: bool = False
    pool: Optional[str] = None


class JobsLogsBody(RequestBody):
    """The request body for the jobs logs endpoint."""
    name: Optional[str] = None
    job_id: Optional[int] = None
    follow: bool = True
    controller: bool = False
    refresh: bool = False
    tail: Optional[int] = None


class RequestCancelBody(RequestBody):
    """The request body for the API request cancellation endpoint."""
    # Kill all requests if request_ids is None.
    request_ids: Optional[List[str]] = None
    user_id: Optional[str] = None


class RequestStatusBody(pydantic.BaseModel):
    """The request body for the API request status endpoint."""
    request_ids: Optional[List[str]] = None
    all_status: bool = False
    limit: Optional[int] = None
    fields: Optional[List[str]] = None


class ServeUpBody(RequestBody):
    """The request body for the serve up endpoint."""
    task: str
    service_name: str

    def to_kwargs(self) -> Dict[str, Any]:
        kwargs = super().to_kwargs()
        dag = common.process_mounts_in_task_on_api_server(self.task,
                                                          self.env_vars,
                                                          workdir_only=False)
        assert len(
            dag.tasks) == 1, ('Must only specify one task in the DAG for '
                              'a service.', dag)
        kwargs['task'] = dag.tasks[0]
        return kwargs


class ServeUpdateBody(RequestBody):
    """The request body for the serve update endpoint."""
    task: str
    service_name: str
    mode: serve.UpdateMode

    def to_kwargs(self) -> Dict[str, Any]:
        kwargs = super().to_kwargs()
        dag = common.process_mounts_in_task_on_api_server(self.task,
                                                          self.env_vars,
                                                          workdir_only=False)
        assert len(
            dag.tasks) == 1, ('Must only specify one task in the DAG for '
                              'a service.', dag)
        kwargs['task'] = dag.tasks[0]
        return kwargs


class ServeDownBody(RequestBody):
    """The request body for the serve down endpoint."""
    service_names: Optional[Union[str, List[str]]]
    all: bool = False
    purge: bool = False


class ServeLogsBody(RequestBody):
    """The request body for the serve logs endpoint."""
    service_name: str
    target: Union[str, serve.ServiceComponent]
    replica_id: Optional[int] = None
    follow: bool = True
    tail: Optional[int] = None


class ServeDownloadLogsBody(RequestBody):
    """The request body for the serve download logs endpoint."""
    service_name: str
    local_dir: str
    targets: Optional[Union[str, serve.ServiceComponent,
                            List[Union[str, serve.ServiceComponent]]]]
    replica_ids: Optional[List[int]] = None
    tail: Optional[int] = None


class ServeStatusBody(RequestBody):
    """The request body for the serve status endpoint."""
    service_names: Optional[Union[str, List[str]]]


class RealtimeGpuAvailabilityRequestBody(RequestBody):
    """The request body for the realtime GPU availability endpoint."""
    context: Optional[str] = None
    name_filter: Optional[str] = None
    quantity_filter: Optional[int] = None
    is_ssh: Optional[bool] = None


class KubernetesNodeInfoRequestBody(RequestBody):
    """The request body for the kubernetes node info endpoint."""
    context: Optional[str] = None


class SlurmNodeInfoRequestBody(RequestBody):
    """The request body for the slurm node info endpoint."""
    slurm_cluster_name: Optional[str] = None


class ListAcceleratorsBody(RequestBody):
    """The request body for the list accelerators endpoint."""
    gpus_only: bool = True
    name_filter: Optional[str] = None
    region_filter: Optional[str] = None
    quantity_filter: Optional[int] = None
    clouds: Optional[Union[List[str], str]] = None
    all_regions: bool = False
    require_price: bool = True
    case_sensitive: bool = True


class ListAcceleratorCountsBody(RequestBody):
    """The request body for the list accelerator counts endpoint."""
    gpus_only: bool = True
    name_filter: Optional[str] = None
    region_filter: Optional[str] = None
    quantity_filter: Optional[int] = None
    clouds: Optional[Union[List[str], str]] = None


class LocalUpBody(RequestBody):
    """The request body for the local up endpoint."""
    gpus: bool = True
    name: Optional[str] = None
    port_start: Optional[int] = None


class LocalDownBody(RequestBody):
    """The request body for the local down endpoint."""
    name: Optional[str] = None


class SSHUpBody(RequestBody):
    """The request body for the SSH up/down endpoints."""
    infra: Optional[str] = None
    cleanup: bool = False


class ServeTerminateReplicaBody(RequestBody):
    """The request body for the serve terminate replica endpoint."""
    service_name: str
    replica_id: int
    purge: bool = False


class KillRequestProcessesBody(RequestBody):
    """The request body for the kill request processes endpoint."""
    request_ids: List[str]


class StreamBody(pydantic.BaseModel):
    """The request body for the stream endpoint."""
    request_id: Optional[str] = None
    log_path: Optional[str] = None
    tail: Optional[int] = None
    plain_logs: bool = True


class JobsDownloadLogsBody(RequestBody):
    """The request body for the jobs download logs endpoint."""
    name: Optional[str]
    job_id: Optional[int]
    refresh: bool = False
    controller: bool = False
    local_dir: str = constants.SKY_LOGS_DIRECTORY


class JobsPoolApplyBody(RequestBody):
    """The request body for the jobs pool apply endpoint."""
    task: Optional[str] = None
    workers: Optional[int] = None
    pool_name: str
    mode: serve.UpdateMode

    def to_kwargs(self) -> Dict[str, Any]:
        kwargs = super().to_kwargs()
        if self.task is not None:
            dag = common.process_mounts_in_task_on_api_server(
                self.task, self.env_vars, workdir_only=False)
            assert len(
                dag.tasks) == 1, ('Must only specify one task in the DAG for '
                                  'a pool.', dag)
            kwargs['task'] = dag.tasks[0]
        else:
            kwargs['task'] = None
        return kwargs


class JobsPoolDownBody(RequestBody):
    """The request body for the jobs pool down endpoint."""
    pool_names: Optional[Union[str, List[str]]]
    all: bool = False
    purge: bool = False


class JobsPoolStatusBody(RequestBody):
    """The request body for the jobs pool status endpoint."""
    pool_names: Optional[Union[str, List[str]]]


class JobsPoolLogsBody(RequestBody):
    """The request body for the jobs pool logs endpoint."""
    pool_name: str
    target: Union[str, serve.ServiceComponent]
    worker_id: Optional[int] = None
    follow: bool = True
    tail: Optional[int] = None


class JobsPoolDownloadLogsBody(RequestBody):
    """The request body for the jobs pool download logs endpoint."""
    pool_name: str
    local_dir: str
    targets: Optional[Union[str, serve.ServiceComponent,
                            List[Union[str, serve.ServiceComponent]]]]
    worker_ids: Optional[List[int]] = None
    tail: Optional[int] = None


class UploadZipFileResponse(pydantic.BaseModel):
    """The response body for the upload zip file endpoint."""
    status: str
    missing_chunks: Optional[List[str]] = None


class UpdateWorkspaceBody(RequestBody):
    """The request body for updating a specific workspace configuration."""
    workspace_name: str = ''  # Will be set from path parameter
    config: Dict[str, Any]


class CreateWorkspaceBody(RequestBody):
    """The request body for creating a new workspace."""
    workspace_name: str = ''  # Will be set from path parameter
    config: Dict[str, Any]


class DeleteWorkspaceBody(RequestBody):
    """The request body for deleting a workspace."""
    workspace_name: str


class UpdateConfigBody(RequestBody):
    """The request body for updating the entire SkyPilot configuration."""
    config: Dict[str, Any]


class GetConfigBody(RequestBody):
    """The request body for getting the entire SkyPilot configuration."""
    pass


class CostReportBody(RequestBody):
    """The request body for the cost report endpoint."""
    days: Optional[int] = 30
    # we use hashes instead of names to avoid the case where
    # the name is not unique
    cluster_hashes: Optional[List[str]] = None
    # Only return fields that are needed for the dashboard
    # summary page
    dashboard_summary_response: bool = False


class RequestPayload(BasePayload):
    """The payload for the requests."""

    request_id: str
    name: str
    entrypoint: str
    request_body: str
    status: str
    created_at: float
    user_id: str
    return_value: str
    error: str
    pid: Optional[int]
    schedule_type: str
    user_name: Optional[str] = None
    # Resources the request operates on.
    cluster_name: Optional[str] = None
    status_msg: Optional[str] = None
    should_retry: bool = False
    finished_at: Optional[float] = None


class SlurmGpuAvailabilityRequestBody(RequestBody):
    """Request body for getting Slurm real-time GPU availability."""
    name_filter: Optional[str] = None
    quantity_filter: Optional[int] = None



""" Shadeform Cloud. """

import json
import os
import typing
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

from sky import catalog
from sky import clouds
from sky.adaptors import common as adaptors_common
from sky.catalog import shadeform_catalog
from sky.utils import registry
from sky.utils import resources_utils
from sky.utils import status_lib

if typing.TYPE_CHECKING:
    from sky import resources as resources_lib
    from sky.utils import volume as volume_lib
else:
    requests = adaptors_common.LazyImport('requests')

# Minimum set of files under ~/.shadeform that grant Shadeform access.
_CREDENTIAL_FILES = [
    'api_key',
]


@registry.CLOUD_REGISTRY.register
class Shadeform(clouds.Cloud):
    """Shadeform GPU Cloud

    Shadeform is a unified API for deploying and managing cloud GPUs across
    multiple cloud providers.
    """

    # Shadeform doesn't have explicit cluster name limits, but conservative
    _MAX_CLUSTER_NAME_LEN_LIMIT = 120

    # Features not currently supported by Shadeform
    # yapf: disable
    _CLOUD_UNSUPPORTED_FEATURES = {
        clouds.CloudImplementationFeatures.STOP:
            'Stopping instances not supported on Shadeform.',
        clouds.CloudImplementationFeatures.MULTI_NODE:
            'Multi-node clusters not supported on Shadeform.',
        clouds.CloudImplementationFeatures.SPOT_INSTANCE:
            'Spot instances not supported on Shadeform.',
        clouds.CloudImplementationFeatures.CUSTOM_DISK_TIER:
            'Custom disk tiers not supported on Shadeform.',
        clouds.CloudImplementationFeatures.CUSTOM_NETWORK_TIER:
            'Custom network tiers not supported on Shadeform.',
        clouds.CloudImplementationFeatures.STORAGE_MOUNTING:
            'Object storage mounting not supported on Shadeform.',
        clouds.CloudImplementationFeatures.HOST_CONTROLLERS:
            'Host controllers not supported on Shadeform.',
        clouds.CloudImplementationFeatures.HIGH_AVAILABILITY_CONTROLLERS:
            'High availability controllers not supported.',
        clouds.CloudImplementationFeatures.CLONE_DISK_FROM_CLUSTER:
            'Disk cloning not supported on Shadeform.',
        clouds.CloudImplementationFeatures.IMAGE_ID:
            'Custom image IDs not supported on Shadeform.',
        clouds.CloudImplementationFeatures.DOCKER_IMAGE:
            'Docker images not supported on Shadeform yet.',
        clouds.CloudImplementationFeatures.CUSTOM_MULTI_NETWORK:
            'Custom multiple network interfaces not supported.',
    }
    # yapf: enable

    _regions: List[clouds.Region] = []

    PROVISIONER_VERSION = clouds.ProvisionerVersion.SKYPILOT
    STATUS_VERSION = clouds.StatusVersion.SKYPILOT
    OPEN_PORTS_VERSION = clouds.OpenPortsVersion.LAUNCH_ONLY

    @classmethod
    def _unsupported_features_for_resources(
        cls,
        resources: 'resources_lib.Resources',
        region: Optional[str] = None,
    ) -> Dict[clouds.CloudImplementationFeatures, str]:
        """The features not supported based on the resources provided."""
        del resources  # unused
        return cls._CLOUD_UNSUPPORTED_FEATURES

    @classmethod
    def _max_cluster_name_length(cls) -> Optional[int]:
        return cls._MAX_CLUSTER_NAME_LEN_LIMIT

    @classmethod
    def regions_with_offering(
        cls,
        instance_type: str,
        accelerators: Optional[Dict[str, int]],
        use_spot: bool,
        region: Optional[str],
        zone: Optional[str],
        resources: Optional['resources_lib.Resources'] = None,
    ) -> List[clouds.Region]:
        """Get regions that offer the requested instance type."""
        assert zone is None, 'Shadeform does not support zones.'
        del zone  # unused
        if use_spot:
            return []  # No spot support

        # IMPORTANT: instance_type here is the specific Shadeform instance type
        # (like 'massedcompute_A6000_base'), NOT the accelerator name
        # We only return regions where this exact instance type exists
        regions = shadeform_catalog.get_region_zones_for_instance_type(
            instance_type, use_spot)

        if region is not None:
            regions = [r for r in regions if r.name == region]
        return regions

    @classmethod
    def zones_provision_loop(
        cls,
        *,
        region: str,
        num_nodes: int,
        instance_type: str,
        accelerators: Optional[Dict[str, int]] = None,
        use_spot: bool = False,
    ) -> Iterator[None]:
        """Iterate over zones for provisioning."""
        del num_nodes  # unused
        if use_spot:
            return

        regions = cls.regions_with_offering(instance_type, accelerators,
                                            use_spot, region, None)
        for r in regions:
            assert r.zones is None, r
            yield r.zones

    @classmethod
    def get_vcpus_mem_from_instance_type(
        cls,
        instance_type: str,
    ) -> Tuple[Optional[float], Optional[float]]:
        """Get vCPUs and memory from instance type."""
        return catalog.get_vcpus_mem_from_instance_type(instance_type,
                                                        clouds='shadeform')

    @classmethod
    def get_accelerators_from_instance_type(
        cls,
        instance_type: str,
    ) -> Optional[Dict[str, Union[int, float]]]:
        """Get accelerator information from instance type."""
        return catalog.get_accelerators_from_instance_type(instance_type,
                                                           clouds='shadeform')

    @classmethod
    def get_default_instance_type(
        cls,
        cpus: Optional[str] = None,
        memory: Optional[str] = None,
        disk_tier: Optional[resources_utils.DiskTier] = None,
        region: Optional[str] = None,
        zone: Optional[str] = None,
    ) -> Optional[str]:
        """Get default instance type."""
        del disk_tier  # Not supported
        return catalog.get_default_instance_type(cpus=cpus,
                                                 memory=memory,
                                                 disk_tier=None,
                                                 region=region,
                                                 zone=zone,
                                                 clouds='shadeform')

    @classmethod
    def get_zone_shell_cmd(cls) -> Optional[str]:
        """Return shell command to get the zone of the instance."""
        return None

    @classmethod
    def get_user_identities(cls) -> Optional[List[List[str]]]:
        """Get user identities for Shadeform."""
        # No user identity support needed
        return None

    def instance_type_exists(self, instance_type: str) -> bool:
        return catalog.instance_type_exists(instance_type, 'shadeform')

    def instance_type_to_hourly_cost(self,
                                     instance_type: str,
                                     use_spot: bool,
                                     region: Optional[str] = None,
                                     zone: Optional[str] = None) -> float:
        """Get hourly cost for instance type."""
        if use_spot:
            raise ValueError('Spot instances are not supported on Shadeform')
        return catalog.get_hourly_cost(instance_type,
                                       use_spot=use_spot,
                                       region=region,
                                       zone=zone,
                                       clouds='shadeform')

    def accelerators_to_hourly_cost(self,
                                    accelerators: Dict[str, int],
                                    use_spot: bool,
                                    region: Optional[str] = None,
                                    zone: Optional[str] = None) -> float:
        """Get hourly cost for accelerators."""
        return 0.0

    def get_egress_cost(self, num_gigabytes: float) -> float:
        """Get egress cost."""
        # No explicit egress pricing from Shadeform API
        return 0.0

    def __repr__(self):
        return 'Shadeform'

    @classmethod
    def get_current_user_identity(cls) -> Optional[str]:
        """Get current user identity."""
        return None

    def make_deploy_resources_variables(
        self,
        resources: 'resources_lib.Resources',
        cluster_name: resources_utils.ClusterName,
        region: 'clouds.Region',
        zones: Optional[List['clouds.Zone']],
        num_nodes: int,
        dryrun: bool = False,
        volume_mounts: Optional[List['volume_lib.VolumeMount']] = None,
    ) -> Dict[str, Any]:
        """Make variables for deployment template."""
        del zones, num_nodes, dryrun, volume_mounts  # unused for Shadeform

        # Get instance type
        r = resources.copy(accelerators=None)
        feasible_resources = self._get_feasible_launchable_resources(r)
        instance_type = feasible_resources.resources_list[0].instance_type

        resources_vars = {}
        if instance_type is not None:
            instance_type_split = instance_type.split('_')
            cloud = instance_type_split[0]
            resources_vars.update({
                'instance_type': instance_type,
                'region': region.name,
                'cloud': cloud,
            })

        # Add accelerator resources for Ray
        accelerators = resources.accelerators
        if accelerators is not None:
            resources_vars['custom_resources'] = json.dumps(accelerators,
                                                            separators=(',',
                                                                        ':'))

        return resources_vars

    def get_credential_file_mounts(self) -> Dict[str, str]:
        """Get credential files that need to be mounted."""
        return {
            f'~/.shadeform/{f}': f'~/.shadeform/{f}' for f in _CREDENTIAL_FILES
        }

    @classmethod
    def get_current_user_identity_str(cls) -> Optional[str]:
        """Get current user identity string."""
        return None

    @classmethod
    def check_credentials(
        cls, cloud_capability: clouds.CloudCapability
    ) -> Tuple[bool, Optional[Union[str, Dict[str, str]]]]:
        """Check if Shadeform credentials are properly configured."""
        del cloud_capability  # unused for Shadeform
        try:
            api_key_path = os.path.expanduser('~/.shadeform/api_key')
            if not os.path.exists(api_key_path):
                return False, (f'Shadeform API key not found. '
                               f'Please save your API key to {api_key_path}')

            # Try to read the API key
            with open(api_key_path, 'r', encoding='utf-8') as f:
                api_key = f.read().strip()

            if not api_key:
                return False, f'Shadeform API key is empty in {api_key_path}'

            return True, None

        except (OSError, IOError) as e:
            return False, f'Error checking Shadeform credentials: {str(e)}'

    def _get_feasible_launchable_resources(
        self, resources: 'resources_lib.Resources'
    ) -> 'resources_utils.FeasibleResources':
        """Get feasible launchable resources."""
        if resources.use_spot:
            return resources_utils.FeasibleResources(
                [], [], 'Spot instances are not supported on Shadeform.')

        if resources.instance_type is not None:
            # Instance type is already specified, validate it
            assert resources.is_launchable(), resources
            fuzzy_candidate_list = [resources.instance_type]
            return resources_utils.FeasibleResources([resources],
                                                     fuzzy_candidate_list, None)

        # Map accelerators to instance types
        def _make_resources(instance_type_list):
            resource_list = []
            for instance_type in instance_type_list:
                r = resources.copy(
                    cloud=Shadeform(),
                    instance_type=instance_type,
                    accelerators=resources.
                    accelerators,  # Keep original accelerators
                    cpus=None,
                    memory=None,
                )
                resource_list.append(r)
            return resource_list

        # Handle accelerator requests
        accelerators = resources.accelerators
        if accelerators is not None:
            # Get the first accelerator type and count
            for accelerator_name, accelerator_count in accelerators.items():
                # Get instance types that provide this accelerator
                func = shadeform_catalog.get_instance_type_for_accelerator
                instance_types, errors = func(accelerator_name,
                                              accelerator_count,
                                              use_spot=resources.use_spot)

                if instance_types:
                    # Create separate resource objects for each instance type
                    # This is crucial: each resource will only be considered
                    # for regions where its specific instance type is available
                    all_resources = []
                    all_candidate_names = []

                    # Create one resource per instance type
                    for instance_type in instance_types:
                        resource = resources.copy(
                            cloud=Shadeform(),
                            instance_type=instance_type,
                            accelerators=resources.accelerators,
                            cpus=None,
                            memory=None,
                        )
                        all_resources.append(resource)
                        all_candidate_names.append(instance_type)

                    return resources_utils.FeasibleResources(
                        all_resources, all_candidate_names, None)
                else:
                    error_msg = (f'No instances available for accelerator '
                                 f'{accelerator_name}')
                    if errors:
                        error_msg += f': {"; ".join(errors)}'
                    return resources_utils.FeasibleResources([], [], error_msg)

            # If accelerator not found in mapping, return error
            return resources_utils.FeasibleResources(
                [], [],
                f'Accelerator {list(accelerators.keys())[0]} not supported.')

        # No accelerators specified, return a default instance type
        if accelerators is None:
            # Return a default instance type
            default_instance_type = Shadeform.get_default_instance_type(
                cpus=resources.cpus,
                memory=resources.memory,
                disk_tier=resources.disk_tier,
                region=resources.region,
                zone=resources.zone)
            if default_instance_type is None:
                # TODO: Add hints to all return values in this method to help
                #  users understand why the resources are not launchable.
                return resources_utils.FeasibleResources([], [], None)
            else:
                return resources_utils.FeasibleResources(
                    _make_resources([default_instance_type]), [], None)

    @classmethod
    def _check_compute_credentials(cls) -> Tuple[bool, Optional[str]]:
        """Check compute credentials."""
        success, msg = cls.check_credentials(clouds.CloudCapability.COMPUTE)
        # Convert return type to match expected signature
        if isinstance(msg, dict):
            msg = str(msg)
        return success, msg

    @classmethod
    def query_status(cls, name: str, tag_filters: Dict[str, str],
                     region: Optional[str], zone: Optional[str],
                     **kwargs) -> List[status_lib.ClusterStatus]:
        """Query cluster status."""
        # For validation purposes, return empty list (no existing clusters)
        # Actual status querying is handled by the provisioner
        return []



"""Constants for Kubernetes provisioning."""

NO_GPU_HELP_MESSAGE = ('If your cluster contains GPUs, make sure '
                       'nvidia.com/gpu resource is available on the nodes and '
                       'the node labels for identifying GPUs '
                       '(e.g., skypilot.co/accelerator) are setup correctly. ')

KUBERNETES_IN_CLUSTER_NAMESPACE_ENV_VAR = 'SKYPILOT_IN_CLUSTER_NAMESPACE'

# Name of kubernetes exec auth wrapper script
SKY_K8S_EXEC_AUTH_WRAPPER = 'sky-kube-exec-wrapper'

# PATH envvar for kubectl exec auth execve
SKY_K8S_EXEC_AUTH_PATH = '$HOME/skypilot-runtime/bin:$HOME/google-cloud-sdk/bin:$PATH'  # pylint: disable=line-too-long

# cache directory for kubeconfig with modified exec auth
SKY_K8S_EXEC_AUTH_KUBECONFIG_CACHE = '~/.sky/generated/kubeconfigs'

# Labels for the Pods created by SkyPilot
TAG_RAY_CLUSTER_NAME = 'ray-cluster-name'
TAG_POD_INITIALIZED = 'skypilot-initialized'
TAG_SKYPILOT_DEPLOYMENT_NAME = 'skypilot-deployment-name'

# Pod phases that are not holding PVCs
PVC_NOT_HOLD_POD_PHASES = ['Succeeded', 'Failed']



"""Dependencies for SkyPilot.

This file is imported by setup.py, so:
- It may not be able to import other skypilot modules, since sys.path may not be
  correct.
- It should not import any dependencies, as they may not be installed yet.
"""
import sys
from typing import Dict, List

clouds_with_ray = ['ibm', 'docker', 'scp']

install_requires = [
    'wheel<0.46.0',  # https://github.com/skypilot-org/skypilot/issues/5153
    'setuptools',  # TODO: match version to pyproject.toml once #5153 is fixed
    'pip',
    'cachetools',
    # NOTE: ray requires click>=7.0.
    # click 8.2.0 has a bug in parsing the command line arguments:
    # https://github.com/pallets/click/issues/2894
    # TODO(aylei): remove this once the bug is fixed in click.
    'click >= 7.0, < 8.2.0',
    'colorama',
    'cryptography',
    # Jinja has a bug in older versions because of the lack of pinning
    # the version of the underlying markupsafe package. See:
    # https://github.com/pallets/jinja/issues/1585
    'jinja2 >= 3.0',
    'jsonschema',
    'networkx',
    'pandas>=1.3.0',
    'pendulum',
    # PrettyTable with version >=2.0.0 is required for the support of
    # `add_rows` method.
    'PrettyTable >= 2.0.0',
    'python-dotenv',
    'rich',
    'tabulate',
    # Light weight requirement, can be replaced with "typing" once
    # we deprecate Python 3.7 (this will take a while).
    'typing_extensions',
    # filelock 3.15.0 or higher is required for async file locking.
    'filelock >= 3.15.0',
    'packaging',
    'psutil',
    'pulp',
    # Cython 3.0 release breaks PyYAML 5.4.*
    # (https://github.com/yaml/pyyaml/issues/601)
    # <= 3.13 may encounter https://github.com/ultralytics/yolov5/issues/414
    'pyyaml > 3.13, != 5.4.*',
    'ijson',
    'orjson',
    'requests',
    # SkyPilot inherits from uvicorn.Server to customize the behavior of
    # uvicorn, so we need to pin uvicorn version to avoid potential break
    # changes.
    # Notes for current version check:
    # - uvicorn 0.33.0 is the latest version that supports Python 3.8
    # - uvicorn 0.36.0 removes setup_event_loop thus breaks SkyPilot's custom
    #   behavior.
    'uvicorn[standard] >=0.33.0, <0.36.0',
    'fastapi',
    # Some pydantic versions are not compatible with ray. Adopted from ray's
    # setup.py:
    # https://github.com/ray-project/ray/blob/ray-2.9.3/python/setup.py#L254
    # We need pydantic>=2.0.0 for API server and client.
    'pydantic!=2.0.*,!=2.1.*,!=2.2.*,!=2.3.*,!=2.4.*,<3,>2',
    # Required for Form data by pydantic
    'python-multipart',
    'aiofiles',
    'httpx',
    'setproctitle',
    'sqlalchemy',
    'psycopg2-binary',
    'aiosqlite',
    'asyncpg',
    # TODO(hailong): These three dependencies should be removed after we make
    # the client-side actually not importing them.
    'casbin',
    'sqlalchemy_adapter',
    # Required for API server metrics
    'prometheus_client>=0.8.0',
    'passlib',
    'bcrypt==4.0.1',
    'pyjwt',
    'gitpython',
    'paramiko',
    'types-paramiko',
    'alembic',
    'aiohttp',
    'anyio',
]

# See requirements-dev.txt for the version of grpc and protobuf
# used to generate the code during development.

# The grpc version at runtime has to be newer than the version
# used to generate the code.
GRPC = 'grpcio>=1.63.0'
# >= 5.26.1 because the runtime version can't be older than the version
# used to generate the code.
# < 7.0.0 because code generated for a major version V will be supported by
# protobuf runtimes of version V and V+1.
# https://protobuf.dev/support/cross-version-runtime-guarantee
PROTOBUF = 'protobuf>=5.26.1, < 7.0.0'

server_dependencies = [
    # TODO: Some of these dependencies are also specified in install_requires,
    # so they are redundant here. We should figure out if they are only needed
    # on the server (should remove from install_requires), or if they are needed
    # on the client (should remove from here).
    'casbin',
    'sqlalchemy_adapter',
    'passlib',
    'pyjwt',
    'aiohttp',
    'anyio',
    GRPC,
    PROTOBUF,
    'aiosqlite',
    'greenlet',
]

local_ray = [
    # Lower version of ray will cause dependency conflict for
    # click/grpcio/protobuf.
    # Ray 2.6.1+ resolved cluster launcher bugs
    # and grpcio issues on Apple Silicon.
    # https://github.com/ray-project/ray/releases/tag/ray-2.6.1
    'ray[default] >= 2.6.1',
]

remote = [
    GRPC,
    PROTOBUF,
]

# NOTE: Change the templates/jobs-controller.yaml.j2 file if any of the
# following packages dependencies are changed.
aws_dependencies = [
    # NOTE: this installs CLI V1. To use AWS SSO (e.g., `aws sso login`), users
    # should instead use CLI V2 which is not pip-installable. See
    # https://docs.aws.amazon.com/cli/latest/userguide/getting-started-install.html.
    'awscli>=1.27.10',
    'botocore>=1.29.10',
    'boto3>=1.26.1',
    # NOTE: required by awscli. To avoid ray automatically installing
    # the latest version.
    'colorama < 0.4.5',
]

# Kubernetes 32.0.0 has an authentication bug:
# https://github.com/kubernetes-client/python/issues/2333
kubernetes_dependencies = [
    'kubernetes>=20.0.0,!=32.0.0',
    'websockets',
    'python-dateutil',
]

# azure-cli cannot be installed normally by uv, so we need to work around it in
# a few places.
AZURE_CLI = 'azure-cli>=2.65.0'

cloud_dependencies: Dict[str, List[str]] = {
    'aws': aws_dependencies,
    # TODO(zongheng): azure-cli is huge and takes a long time to install.
    # Tracked in: https://github.com/Azure/azure-cli/issues/7387
    # azure-identity is needed in node_provider.
    # We need azure-identity>=1.13.0 to enable the customization of the
    # timeout of AzureCliCredential.
    'azure': [
        AZURE_CLI,
        'azure-core>=1.31.0',
        'azure-identity>=1.19.0',
        'azure-mgmt-network>=27.0.0',
        'azure-mgmt-compute>=33.0.0',
        'azure-storage-blob>=12.23.1',
        'msgraph-sdk',
        'msrestazure',
    ],
    # We need google-api-python-client>=2.69.0 to enable 'discardLocalSsd'
    # parameter for stopping instances. Reference:
    # https://github.com/googleapis/google-api-python-client/commit/f6e9d3869ed605b06f7cbf2e8cf2db25108506e6
    'gcp': [
        'google-api-python-client>=2.69.0',
        'google-cloud-storage',
        # see https://github.com/conda/conda/issues/13619
        # see https://github.com/googleapis/google-api-python-client/issues/2554
        'pyopenssl >= 23.2.0, <24.3.0',
    ],
    'ibm': [
        'ibm-cloud-sdk-core',
        'ibm-vpc',
        'ibm-platform-services>=0.48.0',
        'ibm-cos-sdk',
    ] + local_ray,
    'docker': ['docker'] + local_ray,
    'lambda': [],  # No dependencies needed for lambda
    'cloudflare': aws_dependencies,
    'coreweave': aws_dependencies + kubernetes_dependencies,
    'scp': local_ray,
    'oci': ['oci'],
    'kubernetes': kubernetes_dependencies,
    'ssh': kubernetes_dependencies,
    # For the container registry auth api. Reference:
    # https://github.com/runpod/runpod-python/releases/tag/1.6.1
    # RunPod needs a TOML parser to read ~/.runpod/config.toml. On Python 3.11+
    # stdlib provides tomllib; on lower versions we depend on tomli explicitly.
    # Instead of installing tomli conditionally, we install it explicitly.
    # This is because the conditional installation of tomli does not work
    # with controller package installation code.
    'runpod': ['runpod>=1.6.1', 'tomli'],
    'fluidstack': [],  # No dependencies needed for fluidstack
    'cudo': ['cudo-compute>=0.1.10'],
    'paperspace': [],  # No dependencies needed for paperspace
    'primeintellect': [],  # No dependencies needed for primeintellect
    'do': ['pydo>=0.3.0', 'azure-core>=1.24.0', 'azure-common'],
    'vast': ['vastai-sdk>=0.1.12'],
    'vsphere': [
        'pyvmomi==8.0.1.0.2',
        # vsphere-automation-sdk is also required, but it does not have
        # pypi release, which cause failure of our pypi release.
        # https://peps.python.org/pep-0440/#direct-references
        # We have the instruction for its installation in our
        # docs instead.
        # 'vsphere-automation-sdk @ git+https://github.com/vmware/vsphere-automation-sdk-python.git@v8.0.1.0' pylint: disable=line-too-long
    ],
    'nebius': [
        # Nebius requires grpcio and protobuf, so we need to include
        # our constraints here.
        'nebius>=0.3.12',
        GRPC,
        PROTOBUF,
    ] + aws_dependencies,
    'hyperbolic': [],  # No dependencies needed for hyperbolic
    'seeweb': ['ecsapi==0.4.0'],
    'shadeform': [],  # No dependencies needed for shadeform
    'slurm': [],  # No dependencies needed for slurm
}

# Calculate which clouds should be included in the [all] installation.
clouds_for_all = set(cloud_dependencies)

if sys.version_info < (3, 10):
    # Nebius needs python3.10. If python 3.9 [all] will not install nebius
    clouds_for_all.remove('nebius')
    clouds_for_all.remove('seeweb')

if sys.version_info >= (3, 12):
    # The version of ray we use does not work with >= 3.12, so avoid clouds
    # that require ray.
    clouds_for_all -= set(clouds_with_ray)
    # vast requires setuptools==51.1.1 which will not work with python >= 3.12
    # TODO: Remove once https://github.com/vast-ai/vast-sdk/pull/6 is released
    clouds_for_all.remove('vast')

cloud_extras = {
    cloud: dependencies + server_dependencies
    for cloud, dependencies in cloud_dependencies.items()
}

extras_require: Dict[str, List[str]] = {
    # Include server_dependencies with each cloud.
    **cloud_extras,
    'all': list(set().union(*[cloud_extras[cloud] for cloud in clouds_for_all])
               ),
    'remote': remote,
    'server': server_dependencies,
}



"""Utility functions for resources."""
import dataclasses
import enum
import itertools
import json
import math
import typing
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from sky import skypilot_config
from sky.skylet import constants
from sky.utils import common_utils
from sky.utils import registry
from sky.utils import ux_utils

if typing.TYPE_CHECKING:
    from sky import backends
    from sky import resources as resources_lib

_PORT_RANGE_HINT_MSG = ('Invalid port range {}. Please use the format '
                        '"from-to", in which from <= to. e.g. "1-3".')
_PORT_HINT_MSG = ('Invalid port {}. '
                  'Please use a port number between 1 and 65535.')
_DEFAULT_MESSAGE_HANDLE_INITIALIZING = '<initializing>'


class DiskTier(enum.Enum):
    """All disk tiers supported by SkyPilot."""
    LOW = 'low'
    MEDIUM = 'medium'
    HIGH = 'high'
    ULTRA = 'ultra'
    BEST = 'best'

    @classmethod
    def supported_tiers(cls) -> List[str]:
        # 'none' is a special tier for overriding
        # the tier specified in task YAML.
        return [tier.value for tier in cls] + ['none']

    @classmethod
    def cli_help_message(cls) -> str:
        return (
            f'OS disk tier. Could be one of {", ".join(cls.supported_tiers())}'
            f'. If {cls.BEST.value} is specified, use the best possible disk '
            'tier. If none is specified, enforce to use default value and '
            f'override the option in task YAML. Default: {cls.MEDIUM.value}')

    def __le__(self, other: 'DiskTier') -> bool:
        types = list(DiskTier)
        return types.index(self) <= types.index(other)


class NetworkTier(enum.Enum):
    """All network tiers supported by SkyPilot."""
    STANDARD = 'standard'
    BEST = 'best'

    @classmethod
    def supported_tiers(cls) -> List[str]:
        return [tier.value for tier in cls]

    @classmethod
    def cli_help_message(cls) -> str:
        return (
            f'Network tier. Could be one of {", ".join(cls.supported_tiers())}'
            f'. If {cls.BEST.value} is specified, use the best network tier '
            'available on the specified instance. '
            f'Default: {cls.STANDARD.value}')

    @classmethod
    def from_str(cls, tier: str) -> 'NetworkTier':
        if tier not in cls.supported_tiers():
            raise ValueError(f'Invalid network tier: {tier}')
        return cls(tier)

    def __le__(self, other: 'NetworkTier') -> bool:
        types = list(NetworkTier)
        return types.index(self) <= types.index(other)


class StorageType(enum.Enum):
    """Storage type."""
    # Durable network storage, e.g. GCP persistent disks
    NETWORK = 'network'
    # Local instance storage, e.g. GCP local SSDs
    INSTANCE = 'instance'


class DiskAttachMode(enum.Enum):
    """Disk attach mode."""
    READ_ONLY = 'read_only'
    READ_WRITE = 'read_write'


@dataclasses.dataclass
class ClusterName:
    display_name: str
    name_on_cloud: str

    def __repr__(self) -> str:
        return repr(self.display_name)

    def __str__(self) -> str:
        return self.display_name


def check_port_str(port: str) -> None:
    if not port.isdigit():
        with ux_utils.print_exception_no_traceback():
            raise ValueError(_PORT_HINT_MSG.format(port))
    if not 1 <= int(port) <= 65535:
        with ux_utils.print_exception_no_traceback():
            raise ValueError(_PORT_HINT_MSG.format(port))


def check_port_range_str(port_range: str) -> None:
    range_list = port_range.split('-')
    if len(range_list) != 2:
        with ux_utils.print_exception_no_traceback():
            raise ValueError(_PORT_RANGE_HINT_MSG.format(port_range))
    from_port, to_port = range_list
    check_port_str(from_port)
    check_port_str(to_port)
    if int(from_port) > int(to_port):
        with ux_utils.print_exception_no_traceback():
            raise ValueError(_PORT_RANGE_HINT_MSG.format(port_range))


def port_ranges_to_set(ports: Optional[List[str]]) -> Set[int]:
    """Parse a list of port ranges into a set that containing no duplicates.

    For example, ['1-3', '5-7'] will be parsed to {1, 2, 3, 5, 6, 7}.
    """
    if ports is None:
        return set()
    port_set = set()
    for port in ports:
        if port.isdigit():
            check_port_str(port)
            port_set.add(int(port))
        else:
            check_port_range_str(port)
            from_port, to_port = port.split('-')
            port_set.update(range(int(from_port), int(to_port) + 1))
    return port_set


def port_set_to_ranges(port_set: Optional[Set[int]]) -> List[str]:
    """Parse a set of ports into the skypilot ports format.

    This function will group consecutive ports together into a range,
    and keep the rest as is. For example, {1, 2, 3, 5, 6, 7} will be
    parsed to ['1-3', '5-7'].
    """
    if port_set is None:
        return []
    ports: List[str] = []
    # Group consecutive ports together.
    # This algorithm is based on one observation: consecutive numbers
    # in a sorted list will have the same difference with their indices.
    # For example, in [1, 2, 3, 5, 6, 7], difference between value and index
    # is [1, 1, 1, 2, 2, 2], and the consecutive numbers are [1, 2, 3] and
    # [5, 6, 7].
    for _, group in itertools.groupby(enumerate(sorted(port_set)),
                                      lambda x: x[1] - x[0]):
        port = [g[1] for g in group]
        if len(port) == 1:
            ports.append(str(port[0]))
        else:
            ports.append(f'{port[0]}-{port[-1]}')
    return ports


def simplify_ports(ports: List[str]) -> List[str]:
    """Simplify a list of ports.

    For example, ['1-2', '3', '5-6', '7'] will be simplified to ['1-3', '5-7'].
    """
    return port_set_to_ranges(port_ranges_to_set(ports))


def format_resource(resource: 'resources_lib.Resources',
                    simplified_only: bool = False) -> Tuple[str, Optional[str]]:
    resource = resource.assert_launchable()
    is_k8s = str(resource.cloud).lower() == 'kubernetes'
    if resource.accelerators is None or is_k8s or not simplified_only:
        vcpu, mem = resource.cloud.get_vcpus_mem_from_instance_type(
            resource.instance_type)

    elements_simple = []
    elements_full = []

    if resource.accelerators is not None:
        acc, count = list(resource.accelerators.items())[0]
        elements_simple.append(f'gpus={acc}:{count}')
        elements_full.append(f'gpus={acc}:{count}')

    if (resource.accelerators is None or is_k8s):
        if vcpu is not None:
            elements_simple.append(f'cpus={int(vcpu)}')
            elements_full.append(f'cpus={int(vcpu)}')
        if mem is not None:
            elements_simple.append(f'mem={int(mem)}')
            elements_full.append(f'mem={int(mem)}')
    elif not simplified_only:
        if vcpu is not None:
            elements_full.append(f'cpus={int(vcpu)}')
        if mem is not None:
            elements_full.append(f'mem={int(mem)}')

    if not is_k8s:
        instance_type_full = resource.instance_type
        instance_type_simple = common_utils.truncate_long_string(
            instance_type_full, 15)
        elements_simple.append(instance_type_simple)
        elements_full.append(instance_type_full)
    elements_simple.append('...')
    if not simplified_only:
        image_id = resource.image_id
        if image_id is not None:
            if None in image_id:
                elements_full.append(f'image_id={image_id[None]}')
            else:
                elements_full.append(f'image_id={image_id}')
        elements_full.append(f'disk={resource.disk_size}')
        disk_tier = resource.disk_tier
        if disk_tier is not None:
            elements_full.append(f'disk_tier={disk_tier.value}')
        ports = resource.ports
        if ports is not None:
            elements_full.append(f'ports={ports}')

    spot = '[spot]' if resource.use_spot else ''
    resources_str_simple = (
        f'{spot}({"" if not elements_simple else ", ".join(elements_simple)})')
    if simplified_only:
        return resources_str_simple, None
    else:
        resources_str_full = (
            f'{spot}({"" if not elements_full else ", ".join(elements_full)})')
        return resources_str_simple, resources_str_full


def get_readable_resources_repr(
        handle: 'backends.CloudVmRayResourceHandle',
        simplified_only: bool = False) -> Tuple[str, Optional[str]]:
    resource_str_simple, resource_str_full = format_resource(
        handle.launched_resources, simplified_only)
    if not simplified_only:
        assert resource_str_full is not None
    if (handle.launched_nodes is not None and
            handle.launched_resources is not None):
        return (f'{handle.launched_nodes}x{resource_str_simple}',
                None if simplified_only else
                f'{handle.launched_nodes}x{resource_str_full}')
    return (_DEFAULT_MESSAGE_HANDLE_INITIALIZING,
            _DEFAULT_MESSAGE_HANDLE_INITIALIZING)


def make_ray_custom_resources_str(
        resource_dict: Optional[Dict[str, Union[int, float]]]) -> Optional[str]:
    """Convert resources to Ray custom resources format."""
    if resource_dict is None:
        return None
    # Ray does not allow fractional resources, so we need to ceil the values.
    ceiled_dict = {k: math.ceil(v) for k, v in resource_dict.items()}
    return json.dumps(ceiled_dict, separators=(',', ':'))


@dataclasses.dataclass
class FeasibleResources:
    """Feasible resources returned by cloud.

    Used to represent a collection of feasible resources returned by cloud,
    any fuzzy candidates, and optionally a string hint if no feasible resources
    are found.

    Fuzzy candidates example: when the requested GPU is A100:1 but is not
    available in a cloud/region, the fuzzy candidates are results of a fuzzy
    search in the catalog that are offered in the location. E.g.,
    ['A100-80GB:1', 'A100-80GB:2', 'A100-80GB:4', 'A100:8']
    """
    resources_list: List['resources_lib.Resources']
    fuzzy_candidate_list: List[str]
    hint: Optional[str]


def need_to_query_reservations() -> bool:
    """Checks if we need to query reservations from cloud APIs.

    We need to query reservations if:
    - The cloud has specific reservations.
    - The cloud prioritizes reservations over on-demand instances.

    This is useful to skip the potentially expensive reservation query for
    clouds that do not use reservations.
    """
    for cloud_str in registry.CLOUD_REGISTRY.keys():
        cloud_specific_reservations = (
            skypilot_config.get_effective_region_config(
                cloud=cloud_str,
                region=None,
                keys=('specific_reservations',),
                default_value=None))
        cloud_prioritize_reservations = (
            skypilot_config.get_effective_region_config(
                cloud=cloud_str,
                region=None,
                keys=('prioritize_reservations',),
                default_value=False))
        if (cloud_specific_reservations is not None or
                cloud_prioritize_reservations):
            return True
    return False


def make_launchables_for_valid_region_zones(
    launchable_resources: 'resources_lib.Resources',
    override_optimize_by_zone: bool = False,
) -> List['resources_lib.Resources']:
    assert launchable_resources.is_launchable()
    # In principle, all provisioning requests should be made at the granularity
    # of a single zone. However, for on-demand instances, we batch the requests
    # to the zones in the same region in order to leverage the region-level
    # provisioning APIs of AWS and Azure. This way, we can reduce the number of
    # API calls, and thus the overall failover time. Note that this optimization
    # does not affect the user cost since the clouds charge the same prices for
    # on-demand instances in the same region regardless of the zones. On the
    # other hand, for spot instances, we do not batch the requests because the
    # "AWS" spot prices may vary across zones.
    # For GCP, we do not batch the requests because GCP reservation system is
    # zone based. Therefore, price estimation is potentially different across
    # zones.

    # NOTE(woosuk): GCP does not support region-level provisioning APIs. Thus,
    # while we return per-region resources here, the provisioner will still
    # issue the request for one zone at a time.
    # NOTE(woosuk): If we support Azure spot instances, we should batch the
    # requests since Azure spot prices are region-level.
    # TODO(woosuk): Batch the per-zone AWS spot instance requests if they are
    # in the same region and have the same price.
    # TODO(woosuk): A better design is to implement batching at a higher level
    # (e.g., in provisioner or optimizer), not here.
    launchables = []
    regions = launchable_resources.get_valid_regions_for_launchable()
    for region in regions:
        assert launchable_resources.cloud is not None, 'Cloud must be specified'
        optimize_by_zone = (override_optimize_by_zone or
                            launchable_resources.cloud.optimize_by_zone())
        # It is possible that we force the optimize_by_zone but some clouds
        # do not support zone-level provisioning (i.e. Azure). So we check
        # if there is zone-level information in the region first.
        if (region.zones is not None and
            (launchable_resources.use_spot or optimize_by_zone)):
            # Spot instances.
            # Do not batch the per-zone requests.
            for zone in region.zones:
                launchables.append(
                    launchable_resources.copy(region=region.name,
                                              zone=zone.name))
        else:
            # On-demand instances.
            # Batch the requests at the granularity of a single region.
            launchables.append(launchable_resources.copy(region=region.name))
    return launchables


def parse_memory_resource(resource_qty_str: Union[str, int, float],
                          field_name: str,
                          ret_type: type = int,
                          unit: str = 'gb',
                          allow_plus: bool = False,
                          allow_x: bool = False,
                          allow_rounding: bool = False) -> str:
    """Returns memory size in chosen units given a resource quantity string.

    Args:
        resource_qty_str: Resource quantity string
        unit: Unit to convert to
        allow_plus: Whether to allow '+' prefix
        allow_x: Whether to allow 'x' suffix
    """
    assert unit in constants.MEMORY_SIZE_UNITS, f'Invalid unit: {unit}'

    error_msg = (f'"{field_name}" field should be a '
                 f'{constants.MEMORY_SIZE_PATTERN}+?,'
                 f' got {resource_qty_str}')

    resource_str = str(resource_qty_str)

    # Handle plus and x suffixes, x is only used internally for jobs controller
    plus = ''
    if resource_str.endswith('+'):
        if allow_plus:
            resource_str = resource_str[:-1]
            plus = '+'
        else:
            raise ValueError(error_msg)

    x = ''
    if resource_str.endswith('x'):
        if allow_x:
            resource_str = resource_str[:-1]
            x = 'x'
        else:
            raise ValueError(error_msg)

    try:
        # We assume it is already in the wanted units to maintain backwards
        # compatibility
        ret_type(resource_str)
        return f'{resource_str}{plus}{x}'
    except ValueError:
        pass

    resource_str = resource_str.lower()
    for mem_unit, multiplier in constants.MEMORY_SIZE_UNITS.items():
        if resource_str.endswith(mem_unit):
            try:
                value = ret_type(resource_str[:-len(mem_unit)])
                converted = (value * multiplier /
                             constants.MEMORY_SIZE_UNITS[unit])
                if not allow_rounding and ret_type(converted) != converted:
                    raise ValueError(error_msg)
                converted = ret_type(converted)
                return f'{converted}{plus}{x}'
            except ValueError:
                continue

    raise ValueError(error_msg)


def parse_time_minutes(time: str) -> int:
    """Convert a time string to minutes.

    Args:
        time: Time string with optional unit suffix (e.g., '30m', '2h', '1d')

    Returns:
        Time in minutes as an integer
    """
    time_str = str(time)

    if time_str.isdecimal():
        # We assume it is already in minutes to maintain backwards
        # compatibility
        return int(time_str)

    time_str = time_str.lower()
    for unit, multiplier in constants.TIME_UNITS.items():
        if time_str.endswith(unit):
            try:
                value = float(time_str[:-len(unit)])
                final_value = math.ceil(value * multiplier)
                if final_value >= 0:
                    return final_value
            except ValueError:
                continue

    raise ValueError(f'Invalid time format: {time}')


def normalize_any_of_resources_config(
        any_of: List[Dict[str, Any]]) -> Tuple[str, ...]:
    """Normalize a list of any_of resources config to a canonical form.

    Args:
        any_of: A list of any_of resources config.

    Returns:
        A normalized tuple representation that can be compared for equality.
        Two lists with the same resource configurations in different orders
        will produce the same normalized result.
    """
    if not any_of:
        return tuple()

    # Convert each config to JSON string with sorted keys, then sort the list
    normalized_configs = [
        json.dumps(config, sort_keys=True, separators=(',', ':'))
        for config in any_of
    ]

    return tuple(sorted(normalized_configs))



"""YAML utilities."""
import io
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

from sky.adaptors import common

if TYPE_CHECKING:
    import yaml
else:
    yaml = common.LazyImport('yaml')

_c_extension_unavailable = False


def safe_load(stream) -> Any:
    global _c_extension_unavailable
    if _c_extension_unavailable:
        return yaml.load(stream, Loader=yaml.SafeLoader)

    try:
        return yaml.load(stream, Loader=yaml.CSafeLoader)
    except AttributeError:
        _c_extension_unavailable = True
        return yaml.load(stream, Loader=yaml.SafeLoader)


def safe_load_all(stream) -> Any:
    global _c_extension_unavailable
    if _c_extension_unavailable:
        return yaml.load_all(stream, Loader=yaml.SafeLoader)

    try:
        return yaml.load_all(stream, Loader=yaml.CSafeLoader)
    except AttributeError:
        _c_extension_unavailable = True
        return yaml.load_all(stream, Loader=yaml.SafeLoader)


def read_yaml(path: Optional[str]) -> Dict[str, Any]:
    if path is None:
        raise ValueError('Attempted to read a None YAML.')
    with open(path, 'r', encoding='utf-8') as f:
        config = safe_load(f)
    return config


def read_yaml_str(yaml_str: str) -> Dict[str, Any]:
    stream = io.StringIO(yaml_str)
    parsed_yaml = safe_load(stream)
    if not parsed_yaml:
        # Empty dict
        return {}
    return parsed_yaml


def read_yaml_all_str(yaml_str: str) -> List[Dict[str, Any]]:
    stream = io.StringIO(yaml_str)
    config = safe_load_all(stream)
    configs = list(config)
    if not configs:
        # Empty YAML file.
        return [{}]
    return configs


def read_yaml_all(path: str) -> List[Dict[str, Any]]:
    with open(path, 'r', encoding='utf-8') as f:
        return read_yaml_all_str(f.read())


def dump_yaml(path: str,
              config: Union[List[Dict[str, Any]], Dict[str, Any]],
              blank: bool = False) -> None:
    """Dumps a YAML file.

    Args:
        path: the path to the YAML file.
        config: the configuration to dump.
    """
    with open(path, 'w', encoding='utf-8') as f:
        contents = dump_yaml_str(config)
        if blank and isinstance(config, dict) and len(config) == 0:
            # when dumping to yaml, an empty dict will go in as {}.
            contents = ''
        f.write(contents)


def dump_yaml_str(config: Union[List[Dict[str, Any]], Dict[str, Any]]) -> str:
    """Dumps a YAML string.
    Args:
        config: the configuration to dump.
    Returns:
        The YAML string.
    """

    # https://github.com/yaml/pyyaml/issues/127
    class LineBreakDumper(yaml.SafeDumper):

        def write_line_break(self, data=None):
            super().write_line_break(data)
            if len(self.indents) == 1:
                super().write_line_break()

    if isinstance(config, list):
        dump_func = yaml.dump_all  # type: ignore
    else:
        dump_func = yaml.dump  # type: ignore
    return dump_func(config,
                     Dumper=LineBreakDumper,
                     sort_keys=False,
                     default_flow_style=False)



"""Adding columns for the GC time of task logs and controller logs.

Revision ID: 005
Revises: 004
Create Date: 2025-10-20

"""
# pylint: disable=invalid-name
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa

from sky.utils.db import db_utils

# revision identifiers, used by Alembic.
revision: str = '005'
down_revision: Union[str, Sequence[str], None] = '004'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade():
    """Add columns for logs gc."""
    with op.get_context().autocommit_block():
        db_utils.add_column_to_table_alembic('job_info',
                                             'controller_logs_cleaned_at',
                                             sa.Float(),
                                             server_default=None)
        db_utils.add_column_to_table_alembic('spot',
                                             'logs_cleaned_at',
                                             sa.Float(),
                                             server_default=None)


def downgrade():
    """Remove columns for logs gc."""
    pass



"""Nebius cloud adaptor."""
import asyncio
import os
import threading
from typing import Any, Awaitable, List, Optional

from sky import sky_logging
from sky import skypilot_config
from sky.adaptors import common
from sky.utils import annotations
from sky.utils import ux_utils

# Default read timeout for nebius SDK
READ_TIMEOUT = 10

logger = sky_logging.init_logger(__name__)

_loop_lock = threading.Lock()
_loop = None


def _get_event_loop() -> asyncio.AbstractEventLoop:
    """Get event loop for nebius sdk."""
    global _loop

    if _loop is not None:
        return _loop

    with _loop_lock:
        if _loop is None:
            # Create a new event loop in a dedicated thread
            _loop = asyncio.new_event_loop()
            threading.Thread(target=_loop.run_forever, daemon=True).start()

        return _loop


def sync_call(awaitable: Awaitable[Any]) -> Any:
    """Synchronously run an awaitable in coroutine.

    This wrapper is used to workaround:
    https://github.com/nebius/pysdk/issues/76

    Uses a dedicated background event loop to avoid conflicts
    with existing asyncio contexts and prevent BlockingIOError.
    """
    loop = _get_event_loop()
    future = asyncio.run_coroutine_threadsafe(_coro(awaitable), loop)
    return future.result()


async def _coro(awaitable: Awaitable[Any]) -> Any:
    """Wrapper coroutine for awaitable."""
    return await awaitable


def tenant_id_path() -> str:
    return '~/.nebius/NEBIUS_TENANT_ID.txt'


def iam_token_path() -> str:
    return '~/.nebius/NEBIUS_IAM_TOKEN.txt'


def domain_path() -> str:
    return '~/.nebius/NEBIUS_DOMAIN.txt'


def credentials_path() -> str:
    workspace_path = skypilot_config.get_workspace_cloud('nebius').get(
        'credentials_file_path', None)
    if workspace_path is not None:
        return workspace_path
    return _get_default_credentials_path()


def _get_workspace_credentials_path() -> Optional[str]:
    """Get credentials path if explicitly set in workspace config."""
    workspace_cred_path = skypilot_config.get_workspace_cloud('nebius').get(
        'credentials_file_path', None)
    return workspace_cred_path


def _get_default_credentials_path() -> str:
    """Get the default credentials path."""
    return '~/.nebius/credentials.json'


def api_domain() -> Optional[str]:
    domain_in_ws_config = skypilot_config.get_workspace_cloud('nebius').get(
        'domain', None)
    if domain_in_ws_config is not None:
        return domain_in_ws_config
    domain_in_config = skypilot_config.get_effective_region_config(
        cloud='nebius', region=None, keys=('domain',), default_value=None)
    if domain_in_config is not None:
        return domain_in_config
    try:
        with open(os.path.expanduser(domain_path()), encoding='utf-8') as file:
            return file.read().strip()
    except FileNotFoundError:
        return None


DEFAULT_REGION = 'eu-north1'

NEBIUS_PROFILE_NAME = 'nebius'

MAX_RETRIES_TO_DISK_CREATE = 120
MAX_RETRIES_TO_INSTANCE_STOP = 120
MAX_RETRIES_TO_INSTANCE_START = 120
MAX_RETRIES_TO_INSTANCE_READY = 240

MAX_RETRIES_TO_DISK_DELETE = 120
MAX_RETRIES_TO_INSTANCE_WAIT = 120  # Maximum number of retries

POLL_INTERVAL = 5

_IMPORT_ERROR_MESSAGE = ('Failed to import dependencies for Nebius AI Cloud.'
                         'Try pip install "skypilot[nebius]"')

nebius = common.LazyImport(
    'nebius',
    import_error_message=_IMPORT_ERROR_MESSAGE,
    # https://github.com/grpc/grpc/issues/37642 to avoid spam in console
    set_loggers=lambda: os.environ.update({'GRPC_VERBOSITY': 'NONE'}))
boto3 = common.LazyImport('boto3', import_error_message=_IMPORT_ERROR_MESSAGE)
botocore = common.LazyImport('botocore',
                             import_error_message=_IMPORT_ERROR_MESSAGE)

_LAZY_MODULES = (boto3, botocore, nebius)
_session_creation_lock = threading.RLock()
_INDENT_PREFIX = '    '
NAME = 'Nebius'
SKY_CHECK_NAME = 'Nebius (for Nebius Object Storae)'


def request_error():
    # pylint: disable=import-outside-toplevel
    from nebius.aio import service_error
    return service_error.RequestError


def compute():
    # pylint: disable=import-outside-toplevel
    from nebius.api.nebius.compute import v1 as compute_v1
    return compute_v1


def iam():
    # pylint: disable=import-outside-toplevel
    from nebius.api.nebius.iam import v1 as iam_v1
    return iam_v1


def billing():
    # pylint: disable=import-outside-toplevel
    from nebius.api.nebius.billing import v1alpha1 as billing_v1alpha1
    return billing_v1alpha1


def nebius_common():
    # pylint: disable=import-outside-toplevel
    from nebius.api.nebius.common import v1 as common_v1
    return common_v1


def vpc():
    # pylint: disable=import-outside-toplevel
    from nebius.api.nebius.vpc import v1 as vpc_v1
    return vpc_v1


def get_iam_token():
    try:
        with open(os.path.expanduser(iam_token_path()),
                  encoding='utf-8') as file:
            return file.read().strip()
    except FileNotFoundError:
        return None


def is_token_or_cred_file_exist():
    return (os.path.exists(os.path.expanduser(iam_token_path())) or
            os.path.exists(os.path.expanduser(credentials_path())))


def get_tenant_id():
    tenant_id_in_ws_config = skypilot_config.get_workspace_cloud('nebius').get(
        'tenant_id', None)
    if tenant_id_in_ws_config is not None:
        return tenant_id_in_ws_config
    tenant_id_in_config = skypilot_config.get_effective_region_config(
        cloud='nebius', region=None, keys=('tenant_id',), default_value=None)
    if tenant_id_in_config is not None:
        return tenant_id_in_config
    try:
        with open(os.path.expanduser(tenant_id_path()),
                  encoding='utf-8') as file:
            return file.read().strip()
    except FileNotFoundError:
        return None


def sdk():
    """Create the Nebius SDK with the correct credentials.

    The order of priority is:
    1. Credentials file specified in workspace config, if set
    2. IAM token file, if set
    3. Default credentials path
    """
    # 1. Check if credentials path is set in workspace config (highest priority)
    workspace_cred_path = _get_workspace_credentials_path()
    if workspace_cred_path is not None:
        # Check if token is also available and warn
        token = get_iam_token()
        if token is not None:
            logger.warning(
                f'Both workspace credentials file ({workspace_cred_path}) and '
                f'IAM token file ({iam_token_path()}) are available. Using '
                'workspace credentials file.')
        return _sdk(None, workspace_cred_path)

    # 2. Check for IAM token file (second priority)
    token = get_iam_token()
    if token is not None:
        return _sdk(token, None)

    # 3. Fall back to default credentials path (lowest priority)
    default_cred_path = _get_default_credentials_path()
    return _sdk(None, default_cred_path)


@annotations.lru_cache(scope='request')
def _sdk(token: Optional[str], cred_path: Optional[str]):
    # Exactly one of token or cred_path must be provided
    assert (token is None) != (cred_path is None), (token, cred_path)
    if token is not None:
        return nebius.sdk.SDK(credentials=token, domain=api_domain())
    if cred_path is not None:
        return nebius.sdk.SDK(
            credentials_file_name=os.path.expanduser(cred_path),
            domain=api_domain(),
        )
    raise ValueError('Either token or credentials file path must be provided')


def get_nebius_credentials(boto3_session):
    """Gets the Nebius credentials from the boto3 session object.

    Args:
        boto3_session: The boto3 session object.
    Returns:
        botocore.credentials.ReadOnlyCredentials object with the R2 credentials.
    """
    nebius_credentials = boto3_session.get_credentials()
    if nebius_credentials is None:
        with ux_utils.print_exception_no_traceback():
            raise ValueError('Nebius credentials not found. Run '
                             '`sky check` to verify credentials are '
                             'correctly set up.')
    return nebius_credentials.get_frozen_credentials()


# lru_cache() is thread-safe and it will return the same session object
# for different threads.
# Reference: https://docs.python.org/3/library/functools.html#functools.lru_cache # pylint: disable=line-too-long
@annotations.lru_cache(scope='global')
def session():
    """Create an AWS session."""
    # Creating the session object is not thread-safe for boto3,
    # so we add a reentrant lock to synchronize the session creation.
    # Reference: https://github.com/boto/boto3/issues/1592
    # However, the session object itself is thread-safe, so we are
    # able to use lru_cache() to cache the session object.
    with _session_creation_lock:
        session_ = boto3.session.Session(profile_name=NEBIUS_PROFILE_NAME)
    return session_


@annotations.lru_cache(scope='global')
def resource(resource_name: str, **kwargs):
    """Create a Nebius resource.

    Args:
        resource_name: Nebius resource name (e.g., 's3').
        kwargs: Other options.
    """
    # Need to use the resource retrieved from the per-thread session
    # to avoid thread-safety issues (Directly creating the client
    # with boto3.resource() is not thread-safe).
    # Reference: https://stackoverflow.com/a/59635814

    session_ = session()

    return session_.resource(resource_name, **kwargs)


@annotations.lru_cache(scope='global')
def client(service_name: str):
    """Create Nebius client of a certain service.

    Args:
        service_name: Nebius service name (e.g., 's3').
        kwargs: Other options.
    """
    # Need to use the client retrieved from the per-thread session
    # to avoid thread-safety issues (Directly creating the client
    # with boto3.client() is not thread-safe).
    # Reference: https://stackoverflow.com/a/59635814

    session_ = session()

    return session_.client(service_name)


@common.load_lazy_modules(_LAZY_MODULES)
def botocore_exceptions():
    """AWS botocore exception."""
    # pylint: disable=import-outside-toplevel
    from botocore import exceptions
    return exceptions


def get_credential_file_paths() -> List[str]:
    """Get the list of credential file paths based on current configuration."""
    paths = {
        # Always include tenant ID and IAM token paths
        tenant_id_path(),
        iam_token_path(),
    }

    # Add workspace-specific credentials path if set
    workspace_cred_path = _get_workspace_credentials_path()
    if workspace_cred_path is not None:
        paths.add(workspace_cred_path)
    # Always add default path in case it's needed for fallback
    paths.add(_get_default_credentials_path())

    return list(paths)



"""Nebius library wrapper for SkyPilot."""
import time
from typing import Any, Dict, List, Optional
import uuid

from sky import sky_logging
from sky import skypilot_config
from sky.adaptors import nebius
from sky.provision.nebius import constants as nebius_constants
from sky.utils import common_utils
from sky.utils import resources_utils

logger = sky_logging.init_logger(__name__)

POLL_INTERVAL = 5

_MAX_OPERATIONS_TO_FETCH = 1000


def retry(func):
    """Decorator to retry a function."""

    def wrapper(*args, **kwargs):
        """Wrapper for retrying a function."""
        cnt = 0
        while True:
            try:
                return func(*args, **kwargs)
            except nebius.nebius.error.QueryError as e:
                if cnt >= 3:
                    raise
                logger.warning('Retrying for exception: '
                               f'{common_utils.format_exception(e)}.')
                time.sleep(POLL_INTERVAL)

    return wrapper


def get_project_by_region(region: str) -> str:
    service = nebius.iam().ProjectServiceClient(nebius.sdk())
    projects = nebius.sync_call(
        service.list(
            nebius.iam().ListProjectsRequest(parent_id=nebius.get_tenant_id()),
            timeout=nebius.READ_TIMEOUT))

    #  Check is there project if in config
    project_id = skypilot_config.get_effective_region_config(
        cloud='nebius', region=region, keys=('project_id',), default_value=None)
    if project_id is not None:
        return project_id
    for project in projects.items:
        if project.status.region == region:
            return project.metadata.id
    raise Exception(f'No project found for region "{region}".')


def get_or_create_gpu_cluster(name: str, project_id: str, fabric: str) -> str:
    """Creates a GPU cluster.
    https://docs.nebius.com/compute/clusters/gpu
    """
    service = nebius.compute().GpuClusterServiceClient(nebius.sdk())
    try:
        cluster = nebius.sync_call(
            service.get_by_name(nebius.nebius_common().GetByNameRequest(
                parent_id=project_id,
                name=name,
            )))
        cluster_id = cluster.metadata.id
    except nebius.request_error():
        cluster = nebius.sync_call(
            service.create(nebius.compute().CreateGpuClusterRequest(
                metadata=nebius.nebius_common().ResourceMetadata(
                    parent_id=project_id,
                    name=name,
                ),
                spec=nebius.compute().GpuClusterSpec(
                    infiniband_fabric=fabric))))
        cluster_id = cluster.resource_id
    return cluster_id


def delete_cluster(name: str, region: str) -> None:
    """Delete a GPU cluster."""
    project_id = get_project_by_region(region)
    service = nebius.compute().GpuClusterServiceClient(nebius.sdk())
    try:
        cluster = nebius.sync_call(
            service.get_by_name(nebius.nebius_common().GetByNameRequest(
                parent_id=project_id,
                name=name,
            )))
        cluster_id = cluster.metadata.id
        logger.debug(f'Found GPU Cluster : {cluster_id}.')
        nebius.sync_call(
            service.delete(
                nebius.compute().DeleteGpuClusterRequest(id=cluster_id)))
        logger.debug(f'Deleted GPU Cluster : {cluster_id}.')
    except nebius.request_error():
        logger.debug('GPU Cluster does not exist.')


def list_instances(project_id: str) -> Dict[str, Dict[str, Any]]:
    """Lists instances associated with API key."""
    service = nebius.compute().InstanceServiceClient(nebius.sdk())
    page_token = ''
    instances = []
    while True:
        result = nebius.sync_call(
            service.list(nebius.compute().ListInstancesRequest(
                parent_id=project_id,
                page_size=100,
                page_token=page_token,
            ),
                         timeout=nebius.READ_TIMEOUT))
        instances.extend(result.items)
        if not result.next_page_token:  # "" means no more pages
            break
        page_token = result.next_page_token

    instance_dict: Dict[str, Dict[str, Any]] = {}
    for instance in instances:
        info = {}
        info['status'] = instance.status.state.name
        info['name'] = instance.metadata.name
        if instance.status.network_interfaces:
            info['external_ip'] = instance.status.network_interfaces[
                0].public_ip_address.address.split('/')[0]
            info['internal_ip'] = instance.status.network_interfaces[
                0].ip_address.address.split('/')[0]
        instance_dict[instance.metadata.id] = info

    return instance_dict


def stop(instance_id: str) -> None:
    service = nebius.compute().InstanceServiceClient(nebius.sdk())
    nebius.sync_call(
        service.stop(nebius.compute().StopInstanceRequest(id=instance_id)))
    retry_count = 0
    while retry_count < nebius.MAX_RETRIES_TO_INSTANCE_STOP:
        service = nebius.compute().InstanceServiceClient(nebius.sdk())
        instance = nebius.sync_call(
            service.get(nebius.compute().GetInstanceRequest(id=instance_id,)))
        if instance.status.state.name == 'STOPPED':
            break
        time.sleep(POLL_INTERVAL)
        logger.debug(f'Waiting for instance {instance_id} stopping.')
        retry_count += 1

    if retry_count == nebius.MAX_RETRIES_TO_INSTANCE_STOP:
        raise TimeoutError(
            f'Exceeded maximum retries '
            f'({nebius.MAX_RETRIES_TO_INSTANCE_STOP * POLL_INTERVAL}'
            f' seconds) while waiting for instance {instance_id}'
            f' to be stopped.')


def start(instance_id: str) -> None:
    service = nebius.compute().InstanceServiceClient(nebius.sdk())
    nebius.sync_call(
        service.start(nebius.compute().StartInstanceRequest(id=instance_id)))
    retry_count = 0
    while retry_count < nebius.MAX_RETRIES_TO_INSTANCE_START:
        service = nebius.compute().InstanceServiceClient(nebius.sdk())
        instance = nebius.sync_call(
            service.get(nebius.compute().GetInstanceRequest(id=instance_id,)))
        if instance.status.state.name == 'RUNNING':
            break
        time.sleep(POLL_INTERVAL)
        logger.debug(f'Waiting for instance {instance_id} starting.')
        retry_count += 1

    if retry_count == nebius.MAX_RETRIES_TO_INSTANCE_START:
        raise TimeoutError(
            f'Exceeded maximum retries '
            f'({nebius.MAX_RETRIES_TO_INSTANCE_START * POLL_INTERVAL}'
            f' seconds) while waiting for instance {instance_id}'
            f' to be ready.')


def launch(cluster_name_on_cloud: str,
           node_type: str,
           platform: str,
           preset: str,
           region: str,
           image_family: str,
           disk_size: int,
           user_data: str,
           associate_public_ip_address: bool,
           filesystems: List[Dict[str, Any]],
           use_static_ip_address: bool = False,
           use_spot: bool = False,
           network_tier: Optional[resources_utils.NetworkTier] = None) -> str:
    # Each node must have a unique name to avoid conflicts between
    # multiple worker VMs. To ensure uniqueness,a UUID is appended
    # to the node name.
    instance_name = (f'{cluster_name_on_cloud}-'
                     f'{uuid.uuid4().hex[:4]}-{node_type}')
    logger.debug(f'Launching instance: {instance_name}')

    disk_name = 'disk-' + instance_name
    cluster_id = None
    project_id = get_project_by_region(region)
    # 8 GPU virtual machines can be grouped into a GPU cluster.
    # The GPU clusters are built with InfiniBand secure high-speed networking.
    # https://docs.nebius.com/compute/clusters/gpu
    if platform in nebius_constants.INFINIBAND_INSTANCE_PLATFORMS:
        if preset == '8gpu-128vcpu-1600gb':
            fabric = skypilot_config.get_effective_region_config(
                cloud='nebius',
                region=region,
                keys=('fabric',),
                default_value=None)

            # Auto-select fabric if network_tier=best and no fabric configured
            if (fabric is None and
                    str(network_tier) == str(resources_utils.NetworkTier.BEST)):
                try:
                    fabric = nebius_constants.get_default_fabric(
                        platform, region)
                    logger.info(f'Auto-selected InfiniBand fabric {fabric} '
                                f'for {platform} in {region}')
                except ValueError as e:
                    logger.warning(
                        f'InfiniBand fabric auto-selection failed: {e}')

            if fabric is None:
                logger.warning(
                    f'Set up fabric for region {region} in ~/.sky/config.yaml '
                    'to use GPU clusters.')
            else:
                cluster_id = get_or_create_gpu_cluster(cluster_name_on_cloud,
                                                       project_id, fabric)

    service = nebius.compute().DiskServiceClient(nebius.sdk())
    disk = nebius.sync_call(
        service.create(nebius.compute().CreateDiskRequest(
            metadata=nebius.nebius_common().ResourceMetadata(
                parent_id=project_id,
                name=disk_name,
            ),
            spec=nebius.compute().DiskSpec(
                source_image_family=nebius.compute().SourceImageFamily(
                    image_family=image_family),
                size_gibibytes=disk_size,
                type=nebius.compute().DiskSpec.DiskType.NETWORK_SSD,
            ))))
    disk_id = disk.resource_id
    retry_count = 0
    while retry_count < nebius.MAX_RETRIES_TO_DISK_CREATE:
        disk = nebius.sync_call(
            service.get_by_name(nebius.nebius_common().GetByNameRequest(
                parent_id=project_id,
                name=disk_name,
            )))
        if disk.status.state.name == 'READY':
            break
        logger.debug(f'Waiting for disk {disk_name} to be ready.')
        time.sleep(POLL_INTERVAL)
        retry_count += 1

    if retry_count == nebius.MAX_RETRIES_TO_DISK_CREATE:
        raise TimeoutError(
            f'Exceeded maximum retries '
            f'({nebius.MAX_RETRIES_TO_DISK_CREATE * POLL_INTERVAL}'
            f' seconds) while waiting for disk {disk_name}'
            f' to be ready.')

    filesystems_spec = []
    if filesystems:
        for fs in filesystems:
            filesystems_spec.append(nebius.compute().AttachedFilesystemSpec(
                mount_tag=fs['filesystem_mount_tag'],
                attach_mode=nebius.compute().AttachedFilesystemSpec.AttachMode[
                    fs['filesystem_attach_mode']],
                existing_filesystem=nebius.compute().ExistingFilesystem(
                    id=fs['filesystem_id'])))

    service = nebius.vpc().SubnetServiceClient(nebius.sdk())
    sub_net = nebius.sync_call(
        service.list(nebius.vpc().ListSubnetsRequest(parent_id=project_id,)))

    service = nebius.compute().InstanceServiceClient(nebius.sdk())
    logger.debug(f'Creating instance {instance_name} in project {project_id}.')
    try:
        nebius.sync_call(
            service.create(nebius.compute().CreateInstanceRequest(
                metadata=nebius.nebius_common().ResourceMetadata(
                    parent_id=project_id,
                    name=instance_name,
                ),
                spec=nebius.compute().InstanceSpec(
                    gpu_cluster=nebius.compute().InstanceGpuClusterSpec(
                        id=cluster_id,) if cluster_id is not None else None,
                    boot_disk=nebius.compute().AttachedDiskSpec(
                        attach_mode=nebius.compute(
                        ).AttachedDiskSpec.AttachMode.READ_WRITE,
                        existing_disk=nebius.compute().ExistingDisk(
                            id=disk_id)),
                    cloud_init_user_data=user_data,
                    resources=nebius.compute().ResourcesSpec(platform=platform,
                                                             preset=preset),
                    filesystems=filesystems_spec if filesystems_spec else None,
                    network_interfaces=[
                        nebius.compute().NetworkInterfaceSpec(
                            subnet_id=sub_net.items[0].metadata.id,
                            ip_address=nebius.compute().IPAddress(),
                            name='network-interface-0',
                            public_ip_address=nebius.compute().PublicIPAddress(
                                static=use_static_ip_address)
                            if associate_public_ip_address else None,
                        )
                    ],
                    recovery_policy=nebius.compute().InstanceRecoveryPolicy.FAIL
                    if use_spot else None,
                    preemptible=nebius.compute().PreemptibleSpec(
                        priority=1,
                        on_preemption=nebius.compute().PreemptibleSpec.
                        PreemptionPolicy.STOP) if use_spot else None,
                ))))
        instance_id = ''
        retry_count = 0
        while retry_count < nebius.MAX_RETRIES_TO_INSTANCE_READY:
            service = nebius.compute().InstanceServiceClient(nebius.sdk())
            instance = nebius.sync_call(
                service.get_by_name(nebius.nebius_common().GetByNameRequest(
                    parent_id=project_id,
                    name=instance_name,
                )))
            instance_id = instance.metadata.id
            if instance.status.state.name == 'STARTING':
                break

            # All Instances initially have state=STOPPED and reconciling=True,
            # so we need to wait until reconciling is False.
            if instance.status.state.name == 'STOPPED' and \
                    not instance.status.reconciling:
                next_token = ''
                total_operations = 0
                while True:
                    operations_response = nebius.sync_call(
                        service.list_operations_by_parent(
                            nebius.compute().ListOperationsByParentRequest(
                                parent_id=project_id,
                                page_size=100,
                                page_token=next_token,
                            )))
                    total_operations += len(operations_response.operations)
                    for operation in operations_response.operations:
                        # Find the most recent operation for the instance.
                        if operation.resource_id == instance_id:
                            error_msg = operation.description
                            if operation.status:
                                error_msg += f' {operation.status.message}'
                            raise RuntimeError(error_msg)
                    # If we've fetched too many operations, or there are no more
                    # operations to fetch, just raise a generic error.
                    if total_operations > _MAX_OPERATIONS_TO_FETCH or \
                            not operations_response.next_page_token:
                        raise RuntimeError(
                            f'Instance {instance_name} failed to start.')
                    next_token = operations_response.next_page_token
            time.sleep(POLL_INTERVAL)
            logger.debug(
                f'Waiting for instance {instance_name} to start running. '
                f'State: {instance.status.state.name}, '
                f'Reconciling: {instance.status.reconciling}')
            retry_count += 1

        if retry_count == nebius.MAX_RETRIES_TO_INSTANCE_READY:
            raise TimeoutError(
                f'Exceeded maximum retries '
                f'({nebius.MAX_RETRIES_TO_INSTANCE_READY * POLL_INTERVAL}'
                f' seconds) while waiting for instance {instance_name}'
                f' to be ready.')
    except nebius.request_error() as e:
        # Handle ResourceExhausted quota limit error. In this case, we need to
        # clean up the disk as VM creation failed and we can't proceed.
        # It cannot be handled by the caller (provisioner)'s teardown logic,
        # as we cannot retrieve the disk id, after the instance creation
        # fails
        logger.warning(f'Failed to launch instance {instance_name}: {e}')
        service = nebius.compute().DiskServiceClient(nebius.sdk())
        nebius.sync_call(
            service.delete(nebius.compute().DeleteDiskRequest(id=disk_id)))
        logger.debug(f'Disk {disk_id} deleted.')
        raise e
    return instance_id


def remove(instance_id: str) -> None:
    """Terminates the given instance."""
    service = nebius.compute().InstanceServiceClient(nebius.sdk())
    result = nebius.sync_call(
        service.get(nebius.compute().GetInstanceRequest(id=instance_id)))
    disk_id = result.spec.boot_disk.existing_disk.id
    nebius.sync_call(
        service.delete(nebius.compute().DeleteInstanceRequest(id=instance_id)))
    retry_count = 0
    # The instance begins deleting and attempts to delete the disk.
    # Must wait until the disk is unlocked and becomes deletable.
    while retry_count < nebius.MAX_RETRIES_TO_DISK_DELETE:
        try:
            service = nebius.compute().DiskServiceClient(nebius.sdk())
            nebius.sync_call(
                service.delete(nebius.compute().DeleteDiskRequest(id=disk_id)))
            break
        except nebius.request_error():
            logger.debug('Waiting for disk deletion.')
            time.sleep(POLL_INTERVAL)
            retry_count += 1

    if retry_count == nebius.MAX_RETRIES_TO_DISK_DELETE:
        raise TimeoutError(
            f'Exceeded maximum retries '
            f'({nebius.MAX_RETRIES_TO_DISK_DELETE * POLL_INTERVAL}'
            f' seconds) while waiting for disk {disk_id}'
            f' to be deleted.')



"""Utilities for the API server."""

from typing import Optional, Type, TypeVar

import fastapi

from sky.server.requests import payloads
from sky.skylet import constants

_BodyT = TypeVar('_BodyT', bound=payloads.RequestBody)


# TODO(aylei): remove this and disable request body construction at server-side
def build_body_at_server(request: Optional[fastapi.Request],
                         body_type: Type[_BodyT], **data) -> _BodyT:
    """Builds the request body at the server.

    For historical reasons, some handlers mimic a client request body
    at server-side in order to coordinate with the interface of executor.
    This will cause issues where the client info like user identity is not
    respected in these handlers. This function is a helper to build the request
    body at server-side with the auth user overridden.
    """
    request_body = body_type(**data)
    if request is not None:
        auth_user = getattr(request.state, 'auth_user', None)
        if auth_user:
            request_body.env_vars[constants.USER_ID_ENV_VAR] = auth_user.id
            request_body.env_vars[constants.USER_ENV_VAR] = auth_user.name
    return request_body



"""Slurm distributed task executor for SkyPilot.

This module is invoked on each Slurm compute node via:
    srun python -m sky.skylet.executor.slurm --script=... --log-dir=...
"""
import argparse
import json
import os
import pathlib
import socket
import subprocess
import sys
import time

import colorama

from sky.skylet.log_lib import run_bash_command_with_log


def _get_ip_address() -> str:
    """Get the IP address of the current node."""
    ip_result = subprocess.run(['hostname', '-I'],
                               capture_output=True,
                               text=True,
                               check=False)
    return ip_result.stdout.strip().split(
    )[0] if ip_result.returncode == 0 else 'unknown'


def _get_job_node_ips() -> str:
    """Get IPs of all nodes in the current Slurm job."""
    nodelist = os.environ.get('SLURM_JOB_NODELIST', '')
    assert nodelist, 'SLURM_JOB_NODELIST is not set'

    # Expand compressed nodelist (e.g., "node[1-3,5]"
    # -> "node1\nnode2\nnode3\nnode5")
    result = subprocess.run(['scontrol', 'show', 'hostnames', nodelist],
                            capture_output=True,
                            text=True,
                            check=False)
    if result.returncode != 0:
        raise RuntimeError(f'Failed to get hostnames for: {nodelist}')

    hostnames = result.stdout.strip().split('\n')
    ips = []
    for hostname in hostnames:
        try:
            ip = socket.gethostbyname(hostname)
            ips.append(ip)
        except socket.gaierror as e:
            raise RuntimeError('Failed to get IP for hostname: '
                               f'{hostname}') from e

    return '\n'.join(ips)


def main():
    parser = argparse.ArgumentParser(
        description='SkyPilot Slurm task runner for distributed execution')
    parser.add_argument('--script', help='User script (inline, shell-quoted)')
    parser.add_argument('--script-path',
                        help='Path to script file (if too long for inline)')
    parser.add_argument('--env-vars',
                        default='{}',
                        help='JSON-encoded environment variables')
    parser.add_argument('--log-dir',
                        required=True,
                        help='Directory for log files')
    parser.add_argument('--cluster-num-nodes',
                        type=int,
                        required=True,
                        help='Total number of nodes in the cluster')
    parser.add_argument('--cluster-ips',
                        required=True,
                        help='Comma-separated list of cluster node IPs')
    parser.add_argument('--task-name',
                        default=None,
                        help='Task name for single-node log prefix')
    parser.add_argument(
        '--is-setup',
        action='store_true',
        help=
        'Whether this is a setup command (affects logging prefix and filename)')
    parser.add_argument('--alloc-signal-file',
                        help='Path to allocation signal file')
    parser.add_argument('--setup-done-signal-file',
                        help='Path to setup-done signal file')
    args = parser.parse_args()

    assert args.script is not None or args.script_path is not None, (
        'Either '
        '--script or --script-path must be provided')

    # Task rank, different from index of the node in the cluster.
    rank = int(os.environ['SLURM_PROCID'])
    num_nodes = int(os.environ.get('SLURM_NNODES', 1))
    is_single_node_cluster = (args.cluster_num_nodes == 1)

    # Determine node index from IP (like Ray's cluster_ips_to_node_id)
    cluster_ips = args.cluster_ips.split(',')
    ip_addr = _get_ip_address()
    try:
        node_idx = cluster_ips.index(ip_addr)
    except ValueError as e:
        raise RuntimeError(f'IP address {ip_addr} not found in '
                           f'cluster IPs: {cluster_ips}') from e
    node_name = 'head' if node_idx == 0 else f'worker{node_idx}'

    # Log files are written to a shared filesystem, so each node must use a
    # unique filename to avoid collisions.
    if args.is_setup:
        # TODO(kevin): This is inconsistent with other clouds, where it is
        # simply called 'setup.log'. On Slurm that is obviously not possible,
        # since the ~/sky_logs directory is shared by all nodes, so
        # 'setup.log' will be overwritten by other nodes.
        # Perhaps we should apply this naming convention to other clouds.
        log_filename = f'setup-{node_name}.log'
    elif is_single_node_cluster:
        log_filename = 'run.log'
    else:
        log_filename = f'{rank}-{node_name}.log'
    log_path = os.path.join(args.log_dir, log_filename)

    if args.script_path:
        with open(args.script_path, 'r', encoding='utf-8') as f:
            script = f.read()
    else:
        script = args.script

    # Parse env vars and add SKYPILOT environment variables
    env_vars = json.loads(args.env_vars)
    if not args.is_setup:
        # For setup, env vars are set in CloudVmRayBackend._setup.
        env_vars['SKYPILOT_NODE_RANK'] = str(rank)
        env_vars['SKYPILOT_NUM_NODES'] = str(num_nodes)
        env_vars['SKYPILOT_NODE_IPS'] = _get_job_node_ips()

    # Signal file coordination for setup/run synchronization
    # Rank 0 touches the allocation signal to indicate resources acquired
    if args.alloc_signal_file is not None and rank == 0:
        pathlib.Path(args.alloc_signal_file).touch()

    # Wait for setup to complete.
    while args.setup_done_signal_file is not None and not os.path.exists(
            args.setup_done_signal_file):
        time.sleep(0.1)

    # Build log prefix
    # For setup on head: (setup pid={pid})
    # For setup on workers: (setup pid={pid}, ip=1.2.3.4)
    # For single-node cluster: (task_name, pid={pid})
    # For multi-node on head: (head, rank=0, pid={pid})
    # For multi-node on workers: (worker1, rank=1, pid={pid}, ip=1.2.3.4)
    # The {pid} placeholder will be replaced by run_with_log
    if args.is_setup:
        # Setup prefix: head (node_idx=0) shows no IP, workers show IP
        if node_idx == 0:
            prefix = (f'{colorama.Fore.CYAN}(setup pid={{pid}})'
                      f'{colorama.Style.RESET_ALL} ')
        else:
            prefix = (f'{colorama.Fore.CYAN}(setup pid={{pid}}, ip={ip_addr})'
                      f'{colorama.Style.RESET_ALL} ')
    elif is_single_node_cluster:
        # Single-node cluster: use task name
        name_str = args.task_name if args.task_name else 'task'
        prefix = (f'{colorama.Fore.CYAN}({name_str}, pid={{pid}})'
                  f'{colorama.Style.RESET_ALL} ')
    else:
        # Multi-node cluster: head (node_idx=0) shows no IP, workers show IP
        if node_idx == 0:
            prefix = (
                f'{colorama.Fore.CYAN}({node_name}, rank={rank}, pid={{pid}})'
                f'{colorama.Style.RESET_ALL} ')
        else:
            prefix = (f'{colorama.Fore.CYAN}'
                      f'({node_name}, rank={rank}, pid={{pid}}, ip={ip_addr})'
                      f'{colorama.Style.RESET_ALL} ')

    returncode = run_bash_command_with_log(script,
                                           log_path,
                                           env_vars=env_vars,
                                           stream_logs=True,
                                           streaming_prefix=prefix)

    sys.exit(returncode)


if __name__ == '__main__':
    main()



"""Add skylet_ssh_tunnel_metadata to clusters.

Revision ID: 008
Revises: 007
Create Date: 2025-09-09

"""
# pylint: disable=invalid-name
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa

from sky.utils.db import db_utils

# revision identifiers, used by Alembic.
revision: str = '008'
down_revision: Union[str, Sequence[str], None] = '007'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade():
    """Add skylet_ssh_tunnel_metadata column to clusters."""
    with op.get_context().autocommit_block():
        db_utils.add_column_to_table_alembic('clusters',
                                             'skylet_ssh_tunnel_metadata',
                                             sa.LargeBinary(),
                                             server_default=None)


def downgrade():
    """No-op for backward compatibility."""
    pass



"""Slurm Catalog."""

import collections
import re
from typing import Dict, List, Optional, Set, Tuple

from sky import check as sky_check
from sky import clouds as sky_clouds
from sky import sky_logging
from sky.catalog import common
from sky.clouds import cloud
from sky.provision.slurm import utils as slurm_utils
from sky.utils import resources_utils

logger = sky_logging.init_logger(__name__)

_DEFAULT_NUM_VCPUS = 2
_DEFAULT_MEMORY_CPU_RATIO = 1


def instance_type_exists(instance_type: str) -> bool:
    """Check if the given instance type is valid for Slurm."""
    return slurm_utils.SlurmInstanceType.is_valid_instance_type(instance_type)


def get_default_instance_type(cpus: Optional[str] = None,
                              memory: Optional[str] = None,
                              disk_tier: Optional[
                                  resources_utils.DiskTier] = None,
                              region: Optional[str] = None,
                              zone: Optional[str] = None) -> Optional[str]:
    # Delete unused parameters.
    del disk_tier, region, zone

    # Slurm provisions resources via --cpus-per-task and --mem.
    instance_cpus = float(
        cpus.strip('+')) if cpus is not None else _DEFAULT_NUM_VCPUS
    if memory is not None:
        if memory.endswith('+'):
            instance_mem = float(memory[:-1])
        elif memory.endswith('x'):
            instance_mem = float(memory[:-1]) * instance_cpus
        else:
            instance_mem = float(memory)
    else:
        instance_mem = instance_cpus * _DEFAULT_MEMORY_CPU_RATIO
    virtual_instance_type = slurm_utils.SlurmInstanceType(
        instance_cpus, instance_mem).name
    return virtual_instance_type


def list_accelerators(
        gpus_only: bool,
        name_filter: Optional[str],
        region_filter: Optional[str],
        quantity_filter: Optional[int],
        case_sensitive: bool = True,
        all_regions: bool = False,
        require_price: bool = True) -> Dict[str, List[common.InstanceTypeInfo]]:
    """List accelerators in Slurm clusters.

    Returns a dictionary mapping GPU type to a list of InstanceTypeInfo objects.
    """
    return list_accelerators_realtime(gpus_only, name_filter, region_filter,
                                      quantity_filter, case_sensitive,
                                      all_regions, require_price)[0]


def list_accelerators_realtime(
    gpus_only: bool = True,
    name_filter: Optional[str] = None,
    region_filter: Optional[str] = None,
    quantity_filter: Optional[int] = None,
    case_sensitive: bool = True,
    all_regions: bool = False,
    require_price: bool = False,
) -> Tuple[Dict[str, List[common.InstanceTypeInfo]], Dict[str, int], Dict[str,
                                                                          int]]:
    """Fetches real-time accelerator information from the Slurm cluster.

    Uses the `get_slurm_node_info_list` helper function.

    Args:
        gpus_only: If True, only return GPU accelerators.
        name_filter: Regex filter for accelerator names (e.g., 'V100', 'gpu').
        region_filter: Optional filter for Slurm partitions.
        quantity_filter: Minimum number of accelerators required per node.
        case_sensitive: Whether name_filter is case-sensitive.
        all_regions: Unused in Slurm context.
        require_price: Unused in Slurm context.

    Returns:
        A tuple of three dictionaries:
        - qtys_map: Maps GPU type to set of InstanceTypeInfo objects for unique
          counts found per node.
        - total_capacity: Maps GPU type to total count across all nodes.
        - total_available: Maps GPU type to total free count across all nodes.
    """
    del gpus_only, all_regions, require_price

    enabled_clouds = sky_check.get_cached_enabled_clouds_or_refresh(
        cloud.CloudCapability.COMPUTE)
    if not sky_clouds.cloud_in_iterable(sky_clouds.Slurm(), enabled_clouds):
        return {}, {}, {}

    if region_filter is None:
        # Get the first available cluster as default
        all_clusters = slurm_utils.get_all_slurm_cluster_names()
        if not all_clusters:
            return {}, {}, {}
        slurm_cluster = all_clusters[0]
    else:
        slurm_cluster = region_filter

    partition_filter = slurm_utils.get_cluster_default_partition(slurm_cluster)

    # Call the helper function to get node info
    slurm_nodes_info = slurm_utils.slurm_node_info(
        slurm_cluster_name=slurm_cluster)

    if not slurm_nodes_info:
        # Customize error message based on filters
        err_msg = 'No matching GPU nodes found in the Slurm cluster'
        filters_applied = []
        if name_filter:
            filters_applied.append(f'gpu_name={name_filter!r}')
        if quantity_filter:
            filters_applied.append(f'quantity>={quantity_filter}')
        if region_filter:
            filters_applied.append(f'cluster={region_filter!r}')
        if filters_applied:
            err_msg += f' with filters ({", ".join(filters_applied)})'
        err_msg += '.'
        logger.error(
            err_msg)  # Log as error as it indicates no usable resources found
        raise ValueError(err_msg)

    # Aggregate results into the required format
    qtys_map: Dict[str,
                   Set[common.InstanceTypeInfo]] = collections.defaultdict(set)
    total_capacity: Dict[str, int] = collections.defaultdict(int)
    total_available: Dict[str, int] = collections.defaultdict(int)

    for node_info in slurm_nodes_info:
        gpu_type = node_info['gpu_type']
        node_total_gpus = node_info['total_gpus']
        node_free_gpus = node_info['free_gpus']
        partition = node_info['partition']

        # Apply name filter to the determined GPU type
        regex_flags = 0 if case_sensitive else re.IGNORECASE
        if name_filter and not re.match(
                name_filter, gpu_type, flags=regex_flags):
            continue

        # Apply quantity filter (total GPUs on node must meet this)
        if quantity_filter and node_total_gpus < quantity_filter:
            continue

        # Apply partition filter if specified
        # TODO(zhwu): when a node is in multiple partitions, the partition
        # mapping from node to partition does not work.
        # if partition_filter and partition != partition_filter:
        #     continue

        # Create InstanceTypeInfo objects for various GPU counts
        # Similar to Kubernetes, generate powers of 2 up to node_total_gpus
        if node_total_gpus > 0:
            count = 1
            while count <= node_total_gpus:
                instance_info = common.InstanceTypeInfo(
                    instance_type=None,  # Slurm doesn't have instance types
                    accelerator_name=gpu_type,
                    accelerator_count=count,
                    cpu_count=node_info['vcpu_count'],
                    memory=node_info['memory_gb'],
                    price=0.0,  # Slurm doesn't have price info
                    region=partition,  # Use partition as region
                    cloud='slurm',  # Specify cloud as 'slurm'
                    device_memory=0.0,  # No GPU memory info from Slurm
                    spot_price=0.0,  # Slurm doesn't have spot pricing
                )
                qtys_map[gpu_type].add(instance_info)
                count *= 2

            # Add the actual total if it's not already included
            # (e.g., if node has 12 GPUs, include counts 1, 2, 4, 8, 12)
            if count // 2 != node_total_gpus:
                instance_info = common.InstanceTypeInfo(
                    instance_type=None,
                    accelerator_name=gpu_type,
                    accelerator_count=node_total_gpus,
                    cpu_count=node_info['vcpu_count'],
                    memory=node_info['memory_gb'],
                    price=0.0,
                    region=partition,
                    cloud='slurm',
                    device_memory=0.0,
                    spot_price=0.0,
                )
                qtys_map[gpu_type].add(instance_info)

        # Map of GPU type -> total count across all matched nodes
        total_capacity[gpu_type] += node_total_gpus

        # Map of GPU type -> total *free* count across all matched nodes
        total_available[gpu_type] += node_free_gpus

    # Check if any GPUs were found after applying filters
    if not total_capacity:
        err_msg = 'No matching GPU nodes found in the Slurm cluster'
        filters_applied = []
        if name_filter:
            filters_applied.append(f'gpu_name={name_filter!r}')
        if quantity_filter:
            filters_applied.append(f'quantity>={quantity_filter}')
        if partition_filter:
            filters_applied.append(f'partition={partition_filter!r}')
        if filters_applied:
            err_msg += f' with filters ({", ".join(filters_applied)})'
        err_msg += '.'
        logger.error(err_msg)
        raise ValueError(err_msg)

    # Convert sets of InstanceTypeInfo to sorted lists
    final_qtys_map = {
        gpu: sorted(list(instances), key=lambda x: x.accelerator_count)
        for gpu, instances in qtys_map.items()
    }

    logger.debug(f'Aggregated Slurm GPU Info: '
                 f'qtys={final_qtys_map}, '
                 f'capacity={dict(total_capacity)}, '
                 f'available={dict(total_available)}')

    return final_qtys_map, dict(total_capacity), dict(total_available)


def validate_region_zone(
    region_name: Optional[str],
    zone_name: Optional[str],
) -> Tuple[Optional[str], Optional[str]]:
    return (region_name, zone_name)



"""Restarts skylet if version does not match"""

import os
import signal
import subprocess
from typing import List, Optional, Tuple

import psutil

from sky.skylet import constants
from sky.skylet import runtime_utils
from sky.utils import common_utils

VERSION_FILE = runtime_utils.get_runtime_dir_path(constants.SKYLET_VERSION_FILE)
SKYLET_LOG_FILE = runtime_utils.get_runtime_dir_path(constants.SKYLET_LOG_FILE)
PID_FILE = runtime_utils.get_runtime_dir_path(constants.SKYLET_PID_FILE)
PORT_FILE = runtime_utils.get_runtime_dir_path(constants.SKYLET_PORT_FILE)


def _is_running_skylet_process(pid: int) -> bool:
    if pid <= 0:
        return False
    try:
        process = psutil.Process(pid)
        if not process.is_running():
            return False
        # Check if command line contains the skylet module identifier
        cmdline = process.cmdline()
        return any('sky.skylet.skylet' in arg for arg in cmdline)
    except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess,
            OSError) as e:
        print(f'Error checking if skylet process {pid} is running: {e}')
        return False


def _find_running_skylet_pids() -> List[int]:
    if os.path.exists(PID_FILE):
        try:
            with open(PID_FILE, 'r', encoding='utf-8') as pid_file:
                pid = int(pid_file.read().strip())
            if _is_running_skylet_process(pid):
                return [pid]
        except (OSError, ValueError, IOError) as e:
            # Don't fallback to grep-based detection as the existence of the
            # PID file implies that we are on the new version, and there is
            # possibility of there being multiple skylet processes running,
            # and we don't want to accidentally kill the wrong skylet(s).
            print(f'Error reading PID file {PID_FILE}: {e}')
        return []
    else:
        # Fall back to grep-based detection for backward compatibility.
        pids = []
        # We use -m to grep instead of {constants.SKY_PYTHON_CMD} -m to grep
        # because need to handle the backward compatibility of the old skylet
        # started before #3326, which does not use the full path to python.
        proc = subprocess.run(
            'ps aux | grep -v "grep" | grep "sky.skylet.skylet" | grep " -m"',
            shell=True,
            check=False,
            capture_output=True,
            text=True)
        if proc.returncode == 0:
            # Parse the output to extract PIDs (column 2)
            for line in proc.stdout.strip().split('\n'):
                if line:
                    parts = line.split()
                    if len(parts) >= 2:
                        try:
                            pids.append(int(parts[1]))
                        except ValueError:
                            continue
        return pids


def _check_version_match() -> Tuple[bool, Optional[str]]:
    """Check if the version file matches the current skylet version.

    Returns:
        Tuple of (version_match: bool, version: str or None)
    """
    version: Optional[str] = None
    if os.path.exists(VERSION_FILE):
        try:
            with open(VERSION_FILE, 'r', encoding='utf-8') as f:
                version = f.read().strip()
                return version == constants.SKYLET_VERSION, version
        except (OSError, IOError):
            pass
    return False, version


def restart_skylet():
    # Kills old skylet if it is running.
    # TODO(zhwu): make the killing graceful, e.g., use a signal to tell
    # skylet to exit, instead of directly killing it.

    # Find and kill running skylet processes
    for pid in _find_running_skylet_pids():
        try:
            os.kill(pid, signal.SIGKILL)
            # Wait until process fully terminates so its socket gets released.
            # Without this, find_free_port may race with the kernel closing the
            # socket and fail to bind to the port that's supposed to be free.
            psutil.Process(pid).wait(timeout=5)
        except (OSError, ProcessLookupError, psutil.NoSuchProcess,
                psutil.TimeoutExpired):
            # Process died between detection and kill, or timeout waiting
            pass
    # Clean up the PID file
    try:
        os.remove(PID_FILE)
    except OSError:
        pass  # Best effort cleanup

    # TODO(kevin): Handle race conditions here. Race conditions can only
    # happen on Slurm, where there could be multiple clusters running in
    # one network namespace. For other clouds, the behaviour will be that
    # it always gets port 46590 (default port).
    port = common_utils.find_free_port(constants.SKYLET_GRPC_PORT)
    subprocess.run(
        # We have made sure that `attempt_skylet.py` is executed with the
        # skypilot runtime env activated, so that skylet can access the cloud
        # CLI tools.
        f'nohup {constants.SKY_PYTHON_CMD} -m sky.skylet.skylet '
        f'--port={port} '
        f'>> {SKYLET_LOG_FILE} 2>&1 & echo $! > {PID_FILE}',
        shell=True,
        check=True)

    with open(PORT_FILE, 'w', encoding='utf-8') as pf:
        pf.write(str(port))

    with open(VERSION_FILE, 'w', encoding='utf-8') as v_f:
        v_f.write(constants.SKYLET_VERSION)


# Check if our skylet is running
running = bool(_find_running_skylet_pids())

version_match, found_version = _check_version_match()

version_string = (f' (found version {found_version}, new version '
                  f'{constants.SKYLET_VERSION})')
if not running:
    print('Skylet is not running. Starting (version '
          f'{constants.SKYLET_VERSION})...')
elif not version_match:
    print(f'Skylet is stale{version_string}. Restarting...')
else:
    print(
        f'Skylet is running with the latest version {constants.SKYLET_VERSION}.'
    )

if not running or not version_match:
    restart_skylet()



"""A script that generates the Shadeform catalog.

Usage:
    python fetch_shadeform.py [-h] [--api-key API_KEY]
                               [--api-key-path API_KEY_PATH]

If neither --api-key nor --api-key-path are provided, this script will parse
`~/.shadeform/api_key` to look for Shadeform API key.
"""
import argparse
import csv
import json
import os
from typing import Dict

import requests

ENDPOINT = 'https://api.shadeform.ai/v1/instances/types'
DEFAULT_SHADEFORM_API_KEY_PATH = os.path.expanduser('~/.shadeform/api_key')


def parse_gpu_info(gpu_type: str, num_gpus: int, ram_per_gpu: int) -> Dict:
    """Parse GPU information for the catalog."""

    manufacturer = 'NVIDIA'
    if gpu_type == 'MI300X':
        manufacturer = 'AMD'
    elif gpu_type == 'GAUDI2':
        manufacturer = 'Intel'

    return {
        'Gpus': [{
            'Name': gpu_type,
            'Manufacturer': manufacturer,
            'Count': float(num_gpus),
            'MemoryInfo': {
                'SizeInMiB': ram_per_gpu
            },
            'TotalGpuMemoryInMiB': ram_per_gpu * num_gpus
        }]
    }


def create_catalog(api_key: str, output_path: str) -> None:
    """Create Shadeform catalog by fetching from API."""
    headers = {'X-API-KEY': api_key}

    params = {'available': 'true'}

    response = requests.get(ENDPOINT,
                            headers=headers,
                            params=params,
                            timeout=30)
    response.raise_for_status()

    data = response.json()
    instance_types = data.get('instance_types', [])

    with open(output_path, mode='w', encoding='utf-8') as f:
        writer = csv.writer(f, delimiter=',', quotechar='"')
        writer.writerow([
            'InstanceType', 'AcceleratorName', 'AcceleratorCount', 'vCPUs',
            'MemoryGiB', 'Price', 'Region', 'GpuInfo', 'SpotPrice'
        ])

        for instance in instance_types:
            config = instance['configuration']

            cloud = instance['cloud']
            shade_instance_type = instance['shade_instance_type']
            instance_type = f'{cloud}_{shade_instance_type.replace("_", "-")}'
            gpu_type = config['gpu_type'].replace('_', '-')
            gpu_count = float(config['num_gpus'])
            vcpus = float(config['vcpus'])
            memory_gb = int(config['memory_in_gb'])

            # Append "B" to instance_type and gpu_type if they end with "G"
            if instance_type.endswith('G'):
                instance_type += 'B'
            if gpu_type.endswith('G'):
                gpu_type += 'B'

            # Replace "Gx" with "GBx" (case sensitive)
            if 'Gx' in instance_type:
                instance_type = instance_type.replace('Gx', 'GBx')

            # Price is in cents per hour, convert to dollars
            price = float(instance['hourly_price']) / 100

            # Create GPU info
            gpuinfo = None
            if gpu_count > 0:
                gpuinfo_dict = parse_gpu_info(gpu_type, int(gpu_count),
                                              int(config['vram_per_gpu_in_gb']))
                gpuinfo = json.dumps(gpuinfo_dict).replace('"', '\'')

            # Write entry for each available region
            for availability in instance.get('availability', []):
                if availability['available'] and gpu_count > 0:
                    region = availability['region']
                    writer.writerow([
                        instance_type,
                        gpu_type,
                        gpu_count,
                        vcpus,
                        memory_gb,
                        price,
                        region,
                        gpuinfo,
                        ''  # No spot pricing info available
                    ])


def get_api_key(cmdline_args: argparse.Namespace) -> str:
    """Get Shadeform API key from cmdline or default path."""
    api_key = cmdline_args.api_key
    if api_key is None:
        if cmdline_args.api_key_path is not None:
            with open(cmdline_args.api_key_path, mode='r',
                      encoding='utf-8') as f:
                api_key = f.read().strip()
        else:
            # Read from ~/.shadeform/api_key
            with open(DEFAULT_SHADEFORM_API_KEY_PATH,
                      mode='r',
                      encoding='utf-8') as f:
                api_key = f.read().strip()
    assert api_key is not None, (
        f'API key not found. Please provide via --api-key or place in '
        f'{DEFAULT_SHADEFORM_API_KEY_PATH}')
    return api_key


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--api-key', help='Shadeform API key.')
    parser.add_argument('--api-key-path',
                        help='path of file containing Shadeform API key.')
    args = parser.parse_args()
    os.makedirs('shadeform', exist_ok=True)
    create_catalog(get_api_key(args), 'shadeform/vms.csv')
    print('Shadeform catalog saved to shadeform/vms.csv')



"""Load plugins for the SkyPilot API server."""
import abc
import dataclasses
import importlib
import os
from typing import Dict, List, Optional, Tuple

from fastapi import FastAPI

from sky import sky_logging
from sky.skylet import constants as skylet_constants
from sky.utils import common_utils
from sky.utils import config_utils
from sky.utils import yaml_utils

logger = sky_logging.init_logger(__name__)

_DEFAULT_PLUGINS_CONFIG_PATH = '~/.sky/plugins.yaml'
_PLUGINS_CONFIG_ENV_VAR = (
    f'{skylet_constants.SKYPILOT_SERVER_ENV_VAR_PREFIX}PLUGINS_CONFIG')


class ExtensionContext:
    """Context provided to plugins during installation.

    Attributes:
        app: The FastAPI application instance.
        rbac_rules: List of RBAC rules registered by the plugin.
        Example:
        [
            ('user', RBACRule(path='/plugins/api/xx/*', method='POST')),
            ('user', RBACRule(path='/plugins/api/xx/*', method='DELETE'))
        ]
    """

    def __init__(self, app: Optional[FastAPI] = None):
        self.app = app
        self.rbac_rules: List[Tuple[str, RBACRule]] = []

    def register_rbac_rule(self,
                           path: str,
                           method: str,
                           description: Optional[str] = None,
                           role: str = 'user') -> None:
        """Register an RBAC rule for this plugin.

        This method allows plugins to declare which endpoints should be
        restricted to admin users during the install phase.

        Args:
            path: The path pattern to restrict (supports wildcards with
                  keyMatch2).
                  Example: '/plugins/api/credentials/*'
            method: The HTTP method to restrict. Example: 'POST', 'DELETE'
            description: Optional description of what this rule protects.
            role: The role to add this rule to (default: 'user').
                  Rules added to 'user' role block regular users but allow
                  admins.

        Example:
            def install(self, ctx: ExtensionContext):
                # Only admin can upload credentials
                ctx.register_rbac_rule(
                    path='/plugins/api/credentials/*',
                    method='POST',
                    description='Only admin can upload credentials'
                )
        """
        rule = RBACRule(path=path, method=method, description=description)
        self.rbac_rules.append((role, rule))
        logger.debug(f'Registered RBAC rule for {role}: {method} {path}'
                     f'{f" - {description}" if description else ""}')


@dataclasses.dataclass
class RBACRule:
    """RBAC rule for a plugin endpoint.

    Attributes:
        path: The path pattern to match (supports wildcards with keyMatch2).
              Example: '/plugins/api/credentials/*'
        method: The HTTP method to restrict. Example: 'POST', 'DELETE'
        description: Optional description of what this rule protects.
    """
    path: str
    method: str
    description: Optional[str] = None


class BasePlugin(abc.ABC):
    """Base class for all SkyPilot server plugins."""

    @property
    def js_extension_path(self) -> Optional[str]:
        """Optional API route to the JavaScript extension to load."""
        return None

    @abc.abstractmethod
    def install(self, extension_context: ExtensionContext):
        """Hook called by API server to let the plugin install itself."""
        raise NotImplementedError

    def shutdown(self):
        """Hook called by API server to let the plugin shutdown."""
        pass


def _config_schema():
    plugin_schema = {
        'type': 'object',
        'required': ['class'],
        'additionalProperties': False,
        'properties': {
            'class': {
                'type': 'string',
            },
            'parameters': {
                'type': 'object',
                'required': [],
                'additionalProperties': True,
            },
        },
    }
    return {
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            'plugins': {
                'type': 'array',
                'items': plugin_schema,
                'default': [],
            },
        },
    }


def _load_plugin_config() -> Optional[config_utils.Config]:
    """Load plugin config."""
    config_path = os.getenv(_PLUGINS_CONFIG_ENV_VAR,
                            _DEFAULT_PLUGINS_CONFIG_PATH)
    config_path = os.path.expanduser(config_path)
    if not os.path.exists(config_path):
        return None
    config = yaml_utils.read_yaml(config_path) or {}
    common_utils.validate_schema(config,
                                 _config_schema(),
                                 err_msg_prefix='Invalid plugins config: ')
    return config_utils.Config.from_dict(config)


_PLUGINS: Dict[str, BasePlugin] = {}
_EXTENSION_CONTEXT: Optional[ExtensionContext] = None


def load_plugins(extension_context: ExtensionContext):
    """Load and initialize plugins from the config."""
    global _EXTENSION_CONTEXT
    _EXTENSION_CONTEXT = extension_context

    config = _load_plugin_config()
    if not config:
        return

    for plugin_config in config.get('plugins', []):
        class_path = plugin_config['class']
        module_path, class_name = class_path.rsplit('.', 1)
        try:
            module = importlib.import_module(module_path)
        except ImportError as e:
            raise ImportError(
                f'Failed to import plugin module: {module_path}. '
                'Please check if the module is installed in your Python '
                'environment.') from e
        try:
            plugin_cls = getattr(module, class_name)
        except AttributeError as e:
            raise AttributeError(
                f'Could not find plugin {class_name} class in module '
                f'{module_path}. ') from e
        if not issubclass(plugin_cls, BasePlugin):
            raise TypeError(
                f'Plugin {class_path} must inherit from BasePlugin.')
        parameters = plugin_config.get('parameters') or {}
        plugin = plugin_cls(**parameters)
        plugin.install(extension_context)
        _PLUGINS[class_path] = plugin


def get_plugins() -> List[BasePlugin]:
    """Return shallow copies of the registered plugins."""
    return list(_PLUGINS.values())


def get_plugin_rbac_rules() -> Dict[str, List[Dict[str, str]]]:
    """Collect RBAC rules from all loaded plugins.

    Collects rules from the ExtensionContext.

    Returns:
        Dictionary mapping role names to lists of blocklist rules.
        Example:
        {
            'user': [
                {'path': '/plugins/api/credentials/*', 'method': 'POST'},
                {'path': '/plugins/api/credentials/*', 'method': 'DELETE'}
            ]
        }
    """
    rules_by_role: Dict[str, List[Dict[str, str]]] = {}

    # Collect rules registered via ExtensionContext
    if _EXTENSION_CONTEXT:
        for role, rule in _EXTENSION_CONTEXT.rbac_rules:
            if role not in rules_by_role:
                rules_by_role[role] = []
            rules_by_role[role].append({
                'path': rule.path,
                'method': rule.method,
            })

    return rules_by_role



"""RunPod network volume provisioning."""
from typing import Any, Dict, List, Optional, Tuple

from sky import global_user_state
from sky import models
from sky import sky_logging
from sky.adaptors import runpod
from sky.utils import common_utils
from sky.utils import volume as volume_lib

logger = sky_logging.init_logger(__name__)


def _list_volumes() -> List[Dict[str, Any]]:
    # GET /v1/networkvolumes returns a list
    result = runpod.rest_request('GET', '/networkvolumes')
    if isinstance(result, list):
        return result
    # Some deployments may wrap the list.
    if isinstance(result, dict):
        for key in ('items', 'data', 'networkVolumes'):
            if key in result and isinstance(result[key], list):
                return result[key]
    return []


def apply_volume(config: models.VolumeConfig) -> models.VolumeConfig:
    """Create or resolve a RunPod network volume via REST API.

    If a volume with the same `name_on_cloud` exists, reuse it. Otherwise,
    create a new one using POST /v1/networkvolumes.
    """
    name_on_cloud = config.name_on_cloud
    assert name_on_cloud is not None
    data_center_id = config.zone
    if not data_center_id:
        raise RuntimeError('RunPod DataCenterId is required for network '
                           'volumes. Set the zone in the infra field.')
    vol = _try_resolve_volume_by_name(name_on_cloud, data_center_id)
    if vol is None:
        use_existing = config.config.get('use_existing')
        if use_existing:
            raise ValueError(
                f'RunPod network volume {name_on_cloud} does not exist while '
                f'use_existing is True.')
        # Create new volume via REST
        size = config.size
        if size is None:
            raise RuntimeError(
                'RunPod network volume size must be specified to create '
                'a volume.')
        try:
            size_int = int(size)
            if size_int < volume_lib.MIN_RUNPOD_NETWORK_VOLUME_SIZE_GB:
                raise RuntimeError(
                    f'RunPod network volume size must be at least '
                    f'{volume_lib.MIN_RUNPOD_NETWORK_VOLUME_SIZE_GB}GB.')
        except Exception as e:  # pylint: disable=broad-except
            raise RuntimeError(f'Invalid volume size {size!r}: {e}') from e

        payload = {
            'dataCenterId': data_center_id,
            'name': name_on_cloud,
            'size': size_int,
        }
        resp = runpod.rest_request('POST', '/networkvolumes', json=payload)
        if isinstance(resp, dict):
            config.id_on_cloud = resp.get('id')
        else:
            raise RuntimeError(
                f'Failed to create RunPod network volume: {resp}')
        logger.info(f'Created RunPod network volume {name_on_cloud} '
                    f'(id={config.id_on_cloud})')
        return config

    # Use existing matched volume
    id_on_cloud = vol.get('id')
    if id_on_cloud is None:
        raise RuntimeError(
            f'RunPod network volume {name_on_cloud} has no id returned.')
    config.id_on_cloud = id_on_cloud
    size = vol.get('size')
    if size is not None:
        if config.size is not None and config.size != str(size):
            logger.warning(
                f'RunPod network volume {name_on_cloud} has size {size} but '
                f'config size is {config.size}, overriding the config size '
                f'with the volume size.')
        config.size = str(size)
    else:
        logger.warning(
            f'RunPod network volume {name_on_cloud} has no size returned.')
    logger.debug(f'Using existing RunPod network volume {name_on_cloud} '
                 f'(id={config.id_on_cloud})')
    return config


def delete_volume(config: models.VolumeConfig) -> models.VolumeConfig:
    """Deletes a RunPod network volume via REST API if id is known or
       resolvable. If the volume id is not known, try to resolve it by name.
    """
    name_on_cloud = config.name_on_cloud
    assert name_on_cloud is not None
    data_center_id = config.zone
    assert data_center_id is not None
    vol_id = config.id_on_cloud
    if not vol_id:
        vol_id = _try_resolve_volume_id(name_on_cloud, data_center_id)
    if not vol_id:
        logger.warning(
            f'RunPod network volume id not found for {name_on_cloud}; '
            f'skip delete')
        return config
    runpod.rest_request('DELETE', f'/networkvolumes/{vol_id}')
    logger.info(f'Deleted RunPod network volume {name_on_cloud} '
                f'(id={vol_id})')
    return config


def _try_resolve_volume_id(name_on_cloud: str,
                           data_center_id: str) -> Optional[str]:
    vols = _list_volumes()
    matched = next((v for v in vols if v.get('name') == name_on_cloud and
                    v.get('dataCenterId') == data_center_id), None)
    if matched is not None:
        return matched.get('id')
    return None


def _try_resolve_volume_by_name(
        name_on_cloud: str, data_center_id: str) -> Optional[Dict[str, Any]]:
    vols = _list_volumes()
    return next((v for v in vols if v.get('name') == name_on_cloud and
                 v.get('dataCenterId') == data_center_id), None)


def get_volume_usedby(
    config: models.VolumeConfig,) -> Tuple[List[str], List[str]]:
    """Gets the clusters currently using this RunPod network volume.

    Returns:
      (usedby_pods, usedby_clusters)
    usedby_clusters contains SkyPilot cluster display names inferred from
      pod names, which may be wrong.
    """
    vol_id = config.id_on_cloud
    name_on_cloud = config.name_on_cloud
    assert name_on_cloud is not None
    data_center_id = config.zone
    assert data_center_id is not None
    if vol_id is None:
        vol_id = _try_resolve_volume_id(name_on_cloud, data_center_id)
    if vol_id is None:
        return [], []

    # Query all pods for current user and filter by networkVolumeId
    query = """
    query Pods {
      myself {
        pods {
          id
          name
          networkVolumeId
        }
      }
    }
    """
    resp = runpod.runpod.api.graphql.run_graphql_query(query)
    pods = resp.get('data', {}).get('myself', {}).get('pods', [])
    used_pods = [p for p in pods if p.get('networkVolumeId') == vol_id]
    usedby_pod_names = [p.get('name') for p in used_pods if p.get('name')]

    # Map pod names back to SkyPilot cluster names using heuristics.
    clusters = global_user_state.get_clusters()
    cluster_names: List[str] = []
    user_hash = common_utils.get_user_hash()
    for pod_name in usedby_pod_names:
        matched = None
        for c in clusters:
            display = c.get('name')
            if not display:
                continue
            # Heuristic: RunPod pod name is f"{cluster}-{user_hash}-{xxx}"
            # This can be wrong.
            cluster_prefix = display + '-' + user_hash + '-'
            if pod_name.startswith(cluster_prefix):
                matched = display
                break
        if matched and matched not in cluster_names:
            cluster_names.append(matched)

    return usedby_pod_names, cluster_names


def get_all_volumes_usedby(
    configs: List[models.VolumeConfig],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Gets the usedby resources of all volumes."""
    used_by_results = [get_volume_usedby(config) for config in configs]
    used_by_pods, used_by_clusters = {}, {}
    for i in range(len(configs)):
        config = configs[i]
        used_by_pods[config.name_on_cloud] = used_by_results[i][0]
        used_by_clusters[config.name_on_cloud] = used_by_results[i][1]
    return used_by_pods, used_by_clusters


def map_all_volumes_usedby(
        used_by_pods: Dict[str, Any], used_by_clusters: Dict[str, Any],
        config: models.VolumeConfig) -> Tuple[List[str], List[str]]:
    """Maps the usedby resources of a volume."""
    return (used_by_pods.get(config.name_on_cloud,
                             []), used_by_clusters.get(config.name_on_cloud,
                                                       []))



"""Columns for whether the cluster is managed.

Revision ID: 004
Revises: 003
Create Date: 2025-08-07

"""
# pylint: disable=invalid-name
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa

from sky.utils.db import db_utils

# revision identifiers, used by Alembic.π
revision: str = '004'
down_revision: Union[str, Sequence[str], None] = '003'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade():
    """Add columns for whether the cluster is managed."""
    with op.get_context().autocommit_block():
        db_utils.add_column_to_table_alembic('clusters',
                                             'is_managed',
                                             sa.Integer(),
                                             server_default='0')


def downgrade():
    """Remove columns for whether the cluster is managed."""
    pass
